# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module defines and implements several utility functions and language extensions that complement
StrawberryFields. These include:


* **NumPy state functions**

  These functions allow the calculation of various quantum states in either the Fock
  basis (a one-dimensional array indexed by Fock state) or the Gaussian basis (returning
  a vector of means and covariance matrix). These state calculations are NOT done in the
  simulators, but rather in NumPy.

  These are useful for generating states for use in calculating the fidelity of simulations.


* **Random functions**

  These functions generate random numbers and matrices corresponding to various
  quantum states and operations.

* **Decorators**

  The :class:`~.strawberryfields.utils.operation` decorator allows functions
  containing quantum operations acting on a qumode to be used as an
  operation itself within a :class:`.Program` context.

* **Program functions**

  These functions act on :class:`.Program` instances, returning
  or extracting information from the quantum circuit.
"""
import collections
import copy
from inspect import signature

try:
    import tensorflow as tf
except ImportError:
    tf_available = False

import numpy as np
from numpy.polynomial.hermite import hermval
import scipy as sp
from scipy.special import factorial as fac

from .engine import LocalEngine
from .program_utils import Command
from .ops import Gate, Channel, Ket


# ------------------------------------------------------------------------
# State functions - Fock basis and Gaussian basis                |
# ------------------------------------------------------------------------


def squeezed_cov(r, phi, hbar=2):
    r"""Returns the squeezed covariance matrix of a squeezed state

    Args:
        r (complex): the squeezing magnitude
        p (float): the squeezing phase :math:`\phi`
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
    Returns:
        array: the squeezed state
    """
    cov = np.array([[np.exp(-2 * r), 0], [0, np.exp(2 * r)]]) * hbar / 2

    R = np.array([[np.cos(phi / 2), -np.sin(phi / 2)], [np.sin(phi / 2), np.cos(phi / 2)]])

    return np.dot(np.dot(R, cov), R.T)


def vacuum_state(basis="fock", fock_dim=5, hbar=2.0):
    r""" Returns the vacuum state

    Args:
        basis (str): If 'fock', calculates the initial state
            in the Fock basis. If 'gaussian', returns the
            vector of means and the covariance matrix.
        fock_dim (int): the size of the truncated Fock basis if
            using the Fock basis representation
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
    Returns:
        array: the vacuum state
    """
    if basis == "fock":
        state = np.zeros((fock_dim))
        state[0] = 1.0

    elif basis == "gaussian":
        means = np.zeros((2))
        cov = np.identity(2) * hbar / 2
        state = [means, cov]

    return state


def coherent_state(a, basis="fock", fock_dim=5, hbar=2.0):
    r""" Returns the coherent state

    This can be returned either in the Fock basis,

    .. math::

        |\alpha\rangle = e^{-|\alpha|^2/2} \sum_{n=0}^\infty
        \frac{\alpha^n}{\sqrt{n!}}|n\rangle

    or as a Gaussian:

    .. math::

        \mu = (\text{Re}(\alpha),\text{Im}(\alpha)),~~~\sigma = I

    where :math:`\alpha` is the displacement.

    Args:
        a (complex) : the displacement
        basis (str): If 'fock', calculates the initial state
            in the Fock basis. If 'gaussian', returns the
            vector of means and the covariance matrix.
        fock_dim (int): the size of the truncated Fock basis if
            using the Fock basis representation
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
    Returns:
        array: the coherent state
    """
    if basis == "fock":
        state = np.array(
            [np.exp(-0.5 * np.abs(a) ** 2) * a ** n / np.sqrt(fac(n)) for n in range(fock_dim)]
        )

    elif basis == "gaussian":
        means = np.array([a.real, a.imag]) * np.sqrt(2 * hbar)
        cov = np.identity(2) * hbar / 2
        state = [means, cov]

    return state


def squeezed_state(r, p, basis="fock", fock_dim=5, hbar=2.0):
    r""" Returns the squeezed state

    This can be returned either in the Fock basis,

    .. math::

        |z\rangle = \frac{1}{\sqrt{\cosh(r)}}\sum_{n=0}^\infty
        \frac{\sqrt{(2n)!}}{2^n n!}(-e^{i\phi}\tanh(r))^n|2n\rangle

    or as a Gaussian:

    .. math:: \mu = (0,0)

    .. math::
        :nowrap:

        \begin{align*}
            \sigma = R(\phi/2)\begin{bmatrix}e^{-2r} & 0 \\0 & e^{2r} \\\end{bmatrix}R(\phi/2)^T
        \end{align*}

    where :math:`z = re^{i\phi}` is the squeezing factor.

    Args:
        r (complex): the squeezing magnitude
        p (float): the squeezing phase :math:`\phi`
        basis (str): If 'fock', calculates the initial state
            in the Fock basis. If 'gaussian', returns the
            vector of means and the covariance matrix.
        fock_dim (int): the size of the truncated Fock basis if
            using the Fock basis representation
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
    Returns:
        array: the squeezed state
    """
    phi = p

    if basis == "fock":

        def ket(n):
            """Squeezed state kets"""
            return (np.sqrt(fac(2 * n)) / (2 ** n * fac(n))) * (-np.exp(1j * phi) * np.tanh(r)) ** n

        state = np.array([ket(n // 2) if n % 2 == 0 else 0.0 for n in range(fock_dim)])
        state *= np.sqrt(1 / np.cosh(r))

    elif basis == "gaussian":
        means = np.zeros((2))
        state = [means, squeezed_cov(r, phi, hbar)]

    return state


def displaced_squeezed_state(a, r, phi, basis="fock", fock_dim=5, hbar=2.0):
    r""" Returns the squeezed coherent state

    This can be returned either in the Fock basis,

    .. math::

        |\alpha,z\rangle = e^{-\frac{1}{2}|\alpha|^2-\frac{1}{2}{\alpha^*}^2 e^{i\phi}\tanh{(r)}}
        \sum_{n=0}^\infty\frac{\left[\frac{1}{2}e^{i\phi}\tanh(r)\right]^{n/2}}{\sqrt{n!\cosh(r)}}
        H_n\left[ \frac{\alpha\cosh(r)+\alpha^*e^{i\phi}\sinh(r)}{\sqrt{e^{i\phi}\sinh(2r)}} \right]|n\rangle

    where :math:`H_n(x)` is the Hermite polynomial, or as a Gaussian:

    .. math:: \mu = (\text{Re}(\alpha),\text{Im}(\alpha))

    .. math::
        :nowrap:

        \begin{align*}
            \sigma = R(\phi/2)\begin{bmatrix}e^{-2r} & 0 \\0 & e^{2r} \\\end{bmatrix}R(\phi/2)^T
        \end{align*}

    where :math:`z = re^{i\phi}` is the squeezing factor
    and :math:`\alpha` is the displacement.

    Args:
        a (complex): the displacement
        r (complex): the squeezing magnitude
        phi (float): the squeezing phase :math:`\phi`
        basis (str): If 'fock', calculates the initial state
            in the Fock basis. If 'gaussian', returns the
            vector of means and the covariance matrix.
        fock_dim (int): the size of the truncated Fock basis if
            using the Fock basis representation
        hbar (float): (default 2) the value of :math:`\hbar` in the commutation
            relation :math:`[\x,\p]=i\hbar`
    Returns:
        array: the squeezed coherent state
    """
    # pylint: disable=too-many-arguments
    if basis == "fock":

        if r != 0:
            phase_factor = np.exp(1j * phi)
            ch = np.cosh(r)
            sh = np.sinh(r)
            th = np.tanh(r)

            gamma = a * ch + np.conj(a) * phase_factor * sh
            N = np.exp(-0.5 * np.abs(a) ** 2 - 0.5 * np.conj(a) ** 2 * phase_factor * th)

            coeff = np.diag(
                [
                    (0.5 * phase_factor * th) ** (n / 2) / np.sqrt(fac(n) * ch)
                    for n in range(fock_dim)
                ]
            )

            vec = [hermval(gamma / np.sqrt(phase_factor * np.sinh(2 * r)), row) for row in coeff]

            state = N * np.array(vec)

        else:
            state = coherent_state(a, basis="fock", fock_dim=fock_dim)  # pragma: no cover

    elif basis == "gaussian":
        means = np.array([a.real, a.imag]) * np.sqrt(2 * hbar)
        state = [means, squeezed_cov(r, phi, hbar)]

    return state


# ------------------------------------------------------------------------
# State functions - Fock basis only                              |
# ------------------------------------------------------------------------


def fock_state(n, fock_dim=5):
    r""" Returns the Fock state

    Args:
        n (int): the occupation number
        fock_dim (int): the size of the truncated Fock basis
    Returns:
        array: the Fock state
    """
    ket = np.zeros((fock_dim))
    ket[n] = 1.0
    return ket


def cat_state(a, p=0, fock_dim=5):
    r""" Returns the cat state

    .. math::

        |cat\rangle = \frac{1}{\sqrt{2(1+e^{-2|\alpha|^2}\cos(\phi))}}
        \left(|\alpha\rangle +e^{i\phi}|-\alpha\rangle\right)

    with the even cat state given for :math:`\phi=0`, and the odd
    cat state given for :math:`\phi=\pi`.

    Args:
        a (complex): the displacement
        p (float): parity, where :math:`\phi=p\pi`. ``p=0`` corresponds to an even
            cat state, and ``p=1`` an odd cat state
        fock_dim (int): the size of the truncated Fock basis
    Returns:
        array: the cat state
    """
    # p=0 if even, p=pi if odd
    phi = np.pi * p

    # normalisation constant
    temp = np.exp(-0.5 * np.abs(a) ** 2)
    N = temp / np.sqrt(2 * (1 + np.cos(phi) * temp ** 4))

    # coherent states
    k = np.arange(fock_dim)
    c1 = (a ** k) / np.sqrt(fac(k))
    c2 = ((-a) ** k) / np.sqrt(fac(k))

    # add them up with a relative phase
    ket = (c1 + np.exp(1j * phi) * c2) * N

    return ket


# ------------------------------------------------------------------------
# Random numbers and matrices                                           |
# ------------------------------------------------------------------------


def randnc(*arg):
    """Normally distributed array of random complex numbers."""
    return np.random.randn(*arg) + 1j * np.random.randn(*arg)


def random_covariance(N, hbar=2, pure=False, block_diag=False):
    r"""Random covariance matrix.

    Args:
        N (int): number of modes
        hbar (float): the value of :math:`\hbar` to use in the definition
            of the quadrature operators :math:`\x` and :math:`\p`
        pure (bool): If True, a random covariance matrix corresponding
            to a pure state is returned.
        block_diag (bool): If True, uses passive Gaussian transformations that are orthogonal
            instead of unitary. This implies that the positions :math:`q` do not mix with
            the momenta :math:`p` and thus the covariance matrix is block diagonal.
    Returns:
        array: random :math:`2N\times 2N` covariance matrix
    """
    S = random_symplectic(N, block_diag=block_diag)

    if pure:
        return (hbar / 2) * S @ S.T

    nbar = 2 * np.abs(np.random.random(N)) + 1
    Vth = (hbar / 2) * np.diag(np.concatenate([nbar, nbar]))

    return S @ Vth @ S.T


def random_symplectic(N, passive=False, block_diag=False, scale=1.0):
    r"""Random symplectic matrix representing a Gaussian transformation.

    The squeezing parameters :math:`r` for active transformations are randomly
    sampled from the standard normal distribution, while passive transformations
    are randomly sampled from the Haar measure. Note that for the Symplectic
    group there is no notion of Haar measure since this is group is not compact.

    Args:
        N (int): number of modes
        passive (bool): If True, returns a passive Gaussian transformation (i.e.,
            one that preserves photon number). If False (default), returns an active
            transformation.
        block_diag (bool): If True, uses passive Gaussian transformations that are orthogonal
            instead of unitary. This implies that the positions :math:`q` do not mix with
            the momenta :math:`p` and thus the symplectic operator is block diagonal
        scale (float): Sets the scale of the random values used as squeezing parameters.
            They will range from 0 to :math:`\sqrt{2}\texttt{scale}`

    Returns:
        array: random :math:`2N\times 2N` symplectic matrix
    """
    U = random_interferometer(N, real=block_diag)
    O = np.vstack([np.hstack([U.real, -U.imag]), np.hstack([U.imag, U.real])])

    if passive:
        return O

    U = random_interferometer(N, real=block_diag)
    P = np.vstack([np.hstack([U.real, -U.imag]), np.hstack([U.imag, U.real])])

    r = scale * np.abs(randnc(N))
    Sq = np.diag(np.concatenate([np.exp(-r), np.exp(r)]))

    return O @ Sq @ P


def random_interferometer(N, real=False):
    r"""Random unitary matrix representing an interferometer.

    For more details, see :cite:`mezzadri2006`.

    Args:
        N (int): number of modes
        real (bool): return a random real orthogonal matrix

    Returns:
        array: random :math:`N\times N` unitary distributed with the Haar measure
    """
    if real:
        z = np.random.randn(N, N)
    else:
        z = randnc(N, N) / np.sqrt(2.0)
    q, r = sp.linalg.qr(z)
    d = sp.diagonal(r)
    ph = d / np.abs(d)
    U = np.multiply(q, ph, q)
    return U


# ------------------------------------------------------------------------
# Decorators                                                            |
# ------------------------------------------------------------------------


class operation:
    """Groups a sequence of gates into a single operation to be used
    within a Program context.

    For example:

    .. code-block:: python

        @sf.operation(3)
        def custom_operation(v1, v2, q):
            CZgate(v1) | (q[0], q[1])
            Vgate(v2) | q[2]

    Here, the ``operation`` decorator must recieve an argument
    detailing the number of subsystems the resulting custom
    operation acts on.

    The function it acts on can contain arbitrary
    Python and Blackbird code that may normally be placed within a
    Program context. Note that it must always accept the register
    ``q`` it acts on as the *last* argument of the function.

    Once defined, it can be used like any other quantum operation:

    .. code-block:: python

        prog = sf.Program(3)
        with prog.context as q:
            custom_operation(0.5719, 2.0603) | (q[0], q[1], q[3])

    Note that here, we do not pass the register ``q`` directly
    to the function - instead, it is defined on the right hand side
    of the ``|`` operation, like all other Blackbird code.

    Args:
        ns (int): number of subsystems required by the operation
    """

    def __init__(self, ns):
        self.ns = ns
        self.func = None
        self.args = None

    def __or__(self, reg):
        """Apply the operation to a part of a quantum register.

        Redirects the execution flow to the wrapped function.

        Args:
            reg (RegRef, Sequence[RegRef]): subsystem(s) the operation is acting on

        Returns:
            list[RegRef]: subsystem list as RegRefs
        """
        if (not reg) or (not self.ns):
            raise ValueError("Wrong number of subsystems")

        reg_len = 1
        if isinstance(reg, collections.abc.Sized):
            reg_len = len(reg)

        if reg_len != self.ns:
            raise ValueError("Wrong number of subsystems")

        return self._call_function(reg)

    def _call_function(self, reg):
        """Executes the wrapped function and passes the quantum registers.

        Args:
            reg (RegRef, Sequence[RegRef]): subsystem(s) the operation is acting on

        Returns:
            list[RegRef]: subsystem list as RegRefs
        """
        func_sig = signature(self.func)
        num_params = len(func_sig.parameters)

        if num_params == 0:
            raise ValueError("Operation must receive the qumode register as an argument.")

        if num_params != len(self.args) + 1:
            raise ValueError("Mismatch in the number of arguments")

        # pass parameters and subsystems to the function
        if num_params == 1:
            self.func(reg)
        else:
            self.func(*self.args, reg)

        return reg

    def __call__(self, func):
        self.func = func

        def f_proxy(*args):
            """
            Proxy for function execution. Function will actually execute in __or__
            """
            self.args = args
            return self

        return f_proxy


# =================================================
# Program functions
# =================================================


def is_unitary(prog):
    """True iff all the operations in the program are unitary.

    Args:
        prog (Program): quantum program
    Returns:
        bool: True iff all operations in the program are of type :class:`strawberryfields.ops.Gate`
    """
    return all(isinstance(cmd.op, Gate) for cmd in prog.circuit)


def is_channel(prog):
    """True iff all the operations in the program can be represented as quantum channels.

    Args:
        prog (Program): quantum program
    Returns:
        bool: True if all operations in the program are of types :class:`strawberryfields.ops.Gate` and :class:`strawberryfields.ops.Channel`
    """
    # FIXME isn't a preparation also a quantum channel?
    return all(isinstance(cmd.op, (Channel, Gate)) for cmd in prog.circuit)


def _vectorize(tensor):
    """Given a tensor with 4N indices of dimension :math:`D` each, it returns the vectorized
    tensor with 4 indices of dimension :math:`D^N` each. This is the inverse of the procedure
    given by :func:`_unvectorize`.
    Caution: this private method is intended to be used only for Choi and Liouville operators.

    For example, :math:`N=2`,
    ::
        0 --|‾‾‾‾|-- 1
        2 --|    |-- 3
        4 --|    |-- 5
        6 --|____|-- 7

    goes to
    ::
        (0,2) --|‾‾‾‾|-- (1,3)
        (4,6) --|____|-- (5,7)

    Args:
        tensor (array): a tensor with :math:`4N` indices of dimension :math:`D` each

    Returns:
        array: a tensor with 4 indices of dimension :math:`D^N` each

    Raises:
        ValueError: if the input tensor's dimensions are not all equal or if the number
            of its indices is not a multiple of 4
    """
    dims = tensor.ndim

    if dims % 4 != 0:
        raise ValueError(
            "Tensor must have a number of indices that is a multiple of 4, but it has {dims} indices".format(
                dims=dims
            )
        )

    shape = tensor.shape

    if len(set(shape)) != 1:
        raise ValueError(
            "Tensor indices must have all the same dimension, but tensor has shape {shape}".format(
                shape=shape
            )
        )

    transposed = np.einsum(
        tensor, [int(n) for n in np.arange(dims).reshape((2, dims // 2)).T.reshape([-1])]
    )
    vectorized = np.reshape(transposed, [shape[0] ** (dims // 4)] * 4)
    transposed_back = np.einsum("abcd -> acbd", vectorized)

    return transposed_back


def _unvectorize(tensor, num_subsystems):
    """Given a tensor with 4 indices, each of dimension :math:`D^N`, return the unvectorized
    tensor with 4N indices of dimension D each. This is the inverse of the procedure
    given by :func:`_vectorize`.
    Caution: this private method is intended to be used only for Choi and Liouville operators.

    Args:
        tensor (array): a tensor with :math:`4` indices of dimension :math:`D^N`

    Returns:
        array: a tensor with :math:`4N` indices of dimension :math:`D` each

    Raises:
        ValueError: if the input tensor's dimensions are not all equal or if the number
            of its indices is not 4
    """
    dims = tensor.ndim

    if dims != 4:
        raise ValueError("tensor must have 4 indices, but it has {dims} indices".format(dims=dims))

    shape = tensor.shape

    if len(set(shape)) != 1:
        raise ValueError(
            "tensor indices must have all the same dimension, but tensor has shape {shape}".format(
                shape=shape
            )
        )

    transposed = np.einsum("abcd -> acbd", tensor)
    unvectorized = np.reshape(
        transposed, [int(shape[0] ** (1 / num_subsystems))] * (4 * num_subsystems)
    )
    transposed_back = np.einsum(
        unvectorized,
        [
            int(n)
            for n in np.arange(4 * num_subsystems).reshape((2 * num_subsystems, 2)).T.reshape([-1])
        ],
    )

    return transposed_back


def _interleaved_identities(n: int, cutoff_dim: int):
    r"""Maximally entangled state of `n` modes.

    Returns the tensor :math:`\sum_{abc\ldots} \ket{abc\ldots}\bra{abc\ldots}`
    representing an unnormalized, maximally entangled state of `n` subsystems.

    Args:
        n (int): number of subsystems
        cutoff_dim (int): Fock basis truncation dimension

    Returns:
        array: unnormalized maximally entangled state, shape == (cutoff_dim,) * (2*n)
    """
    I = np.identity(cutoff_dim)
    temp = I
    for _ in range(1, n):
        temp = np.tensordot(temp, I, axes=0)

    # use einsum to permute the indices such that |a><a|*|b><b|*|c><c|*... becomes |abc...><abc...|
    sublist = [int(n) for n in np.arange(2 * n).reshape((2, n)).T.reshape([-1])]
    return np.einsum(temp, sublist)


def _program_in_CJ_rep(prog, cutoff_dim: int):
    """Convert a Program object to Choi-Jamiolkowski representation.

    Doubles the number of modes of a Program object and prepends to its circuit
    the preparation of the maximally entangled ket state.

    The core idea is that when we apply any quantum channel (e.g. a unitary gate)
    to the density matrix of the maximally entangled state, we obtain the Choi matrix
    of the channel as the result.

    If the channel is unitary, applying it on the maximally entangled ket yields
    the corresponding unitary matrix, reshaped.

    Args:
        prog (Program): quantum program
        cutoff_dim (int): the Fock basis truncation

    Returns:
        Program: modified program
    """
    prog = copy.deepcopy(prog)
    prog.locked = False  # unlock the copy so we can modify it
    N = prog.init_num_subsystems
    prog._add_subsystems(N)  # pylint: disable=protected-access
    prog.init_num_subsystems = 2 * N
    I = _interleaved_identities(N, cutoff_dim)
    # prepend the circuit with the I ket preparation
    prog.circuit.insert(0, Command(Ket(I), list(prog.reg_refs.values())))
    return prog


def extract_unitary(prog, cutoff_dim: int, vectorize_modes: bool = False, backend: str = "fock"):
    r"""Numerical array representation of a unitary quantum circuit.

    Note that the circuit must only include operations of the :class:`strawberryfields.ops.Gate` class.

    * If ``vectorize_modes=True``, it returns a matrix.
    * If ``vectorize_modes=False``, it returns an operator with :math:`2N` indices,
      where N is the number of modes that the Program is created with. Adjacent
      indices correspond to output-input pairs of the same mode.


    **Example:**

    This shows the Hong-Ou-Mandel effect by extracting the unitary of a 50/50 beamsplitter, and then
    computing the output given by one photon at each input (notice the order of the indices: :math:`[out_1, in_1, out_2, in_2,\dots]`).
    The result tells us that the two photons always emerge together from a random output port and never one per port.

    >>> prog = sf.Program(num_subsystems=2)
    >>> with prog.context as q:
    >>>     BSgate(np.pi/4) | q
    >>> U = extract_unitary(prog, cutoff_dim=3)
    >>> print(abs(U[:,1,:,1])**2)
    [[0.  0.  0.5]
     [0.  0.  0. ]
     [0.5 0.  0. ]])

    Args:
        prog (Program): quantum program
        cutoff_dim (int): dimension of each index
        vectorize_modes (bool): If True, reshape input and output modes in order to return a matrix.
        backend (str): the backend to build the unitary; ``'fock'`` (default) and ``'tf'`` are supported

    Returns:
        array, tf.Tensor: numerical array of the unitary circuit
            as a NumPy ndarray (``'fock'`` backend) or as a TensorFlow Tensor (``'tf'`` backend)

    Raises:
        TypeError: if the operations used to construct the circuit are not all unitary
    """

    if not is_unitary(prog):
        raise TypeError("The circuit definition contains elements that are not of type Gate")

    if backend not in ("fock", "tf"):
        raise ValueError("Only 'fock' and 'tf' backends are supported")

    N = prog.init_num_subsystems
    # extract the unitary matrix by running a modified version of the Program
    p = _program_in_CJ_rep(prog, cutoff_dim)
    eng = LocalEngine(backend, backend_options={"cutoff_dim": cutoff_dim, "pure": True})
    result = eng.run(p).state.ket()

    if vectorize_modes:
        if backend == "fock":
            reshape = np.reshape
        else:
            reshape = tf.reshape
        return reshape(result, [cutoff_dim ** N, cutoff_dim ** N])

    # here we rearrange the indices to go back to the order [in1, out1, in2, out2, etc...]
    if backend == "fock":
        tp = np.transpose
    else:
        tp = tf.transpose
    return tp(result, [int(n) for n in np.arange(2 * N).reshape((2, N)).T.reshape([-1])])


def extract_channel(
    prog, cutoff_dim: int, representation: str = "choi", vectorize_modes: bool = False
):
    r"""Numerical array representation of the channel corresponding to a quantum circuit.

    The representation choices include the Choi state representation, the Liouville representation, and
    the Kraus representation.

    .. note:: Channel extraction can currently only be performed using the ``'fock'`` backend.

    **Tensor shapes**

    * If ``vectorize_modes=True``:

      - ``representation='choi'`` and ``representation='liouville'`` return an array
        with 4 indices
      - ``representation='kraus'`` returns an array of Kraus operators in matrix form


    * If ``vectorize_modes=False``:

      - ``representation='choi'`` and ``representation='liouville'`` return an array
        with :math:`4N` indices
      - ``representation='kraus'`` returns an array of Kraus operators with :math:`2N` indices each,
        where :math:`N` is the number of modes that the Program is created with

    Note that the Kraus representation automatically returns only the non-zero Kraus operators.
    One can reduce the number of operators by discarding Kraus operators with small norm (thus approximating the channel).

    **Choi representation**

    Mathematically, the Choi representation of a channel is a bipartite state :math:`\Lambda_{AB}`
    which contains a complete description of the channel. The way we use it to compute the action
    of the channel :math:`\mathcal{C}` on an input state :math:`\mathcal{\rho}` is as follows:

    .. math::

            \mathcal{C}(\rho) = \mathrm{Tr}[(\rho_A^T\otimes\mathbb{1}_B)\Lambda_{AB}]

    The indices of the non-vectorized Choi operator match exactly those of the state, so that the action
    of the channel can be computed as (e.g., for one mode or for ``vectorize_modes=True``):

    >>> rho_out = np.einsum('ab,abcd', rho_in, choi)

    Notice that this respects the transpose operation.

    For two modes:

    >>> rho_out = np.einsum('abcd,abcdefgh', rho_in, choi)

    Combining consecutive channels (in the order :math:`1,2,3,\dots`) is also straightforward with the Choi operator:

    >>> choi_combined = np.einsum('abcd,cdef,efgh', choi_1, choi_2, choi_3)

    **Liouville operator**

    The Liouville operator is a partial transpose of the Choi operator, such that the first half of
    consecutive index pairs are the output-input right modes (i.e., acting on the "bra" part of the state)
    and the second half are the output-input left modes (i.e., acting on the "ket" part of the state).

    Therefore, the action of the Liouville operator (e.g., for one mode or for ``vectorize_modes=True``) is

    .. math::

            \mathcal{C}(\rho) = \mathrm{unvec}[\mathcal{L}\mathrm{vec}(\rho)]

    where vec() and unvec() are the operations that stack the columns of a matrix to form
    a vector and vice versa.
    In code:

    >>> rho_out = np.einsum('abcd,bd->ca', liouville, rho_in)

    Notice that the state contracts with the second index of each pair and that we output the ket
    on the left (``c``) and the bra on the right (``a``).

    For two modes we have:

    >>> rho_out = np.einsum('abcdefgh,fbhd->eagc', liouville, rho_in)

    The Liouville representation has the property that if the channel is unitary, the operator is separable.
    On the other hand, even if the channel were the identity, the Choi operator would correspond to a maximally entangled state.

    The choi and liouville operators in matrix form (i.e., with two indices) can be found as follows, where
    ``D`` is the dimension of each vectorized index (i.e., for :math:`N` modes, ``D=cutoff_dim**N``):

    >>> choi_matrix = liouville.reshape(D**2, D**2).T
    >>> liouville_matrix = choi.reshape(D**2, D**2).T

    **Kraus representation**

    The Kraus representation is perhaps the most well known:

    .. math::

            \mathcal{C}(\rho) = \sum_k A_k\rho A_k^\dagger

    So to define a channel in the Kraus representation one needs to supply a list of Kraus operators :math:`\{A_k\}`.
    In fact, the result of ``extract_channel`` in the Kraus representation is a rank-3 tensor, where the first
    index is the one indexing the list of operators.

    Adjacent indices of each Kraus operator correspond to output-input pairs of the same mode, so the action
    of the channel can be written as (here for one mode or for ``vectorize_modes=True``):

    >>> rho_out = np.einsum('abc,cd,aed->be', kraus, rho_in, np.conj(kraus))

    Notice the transpose on the third index string (``aed`` rather than ``ade``), as the last operator should be the
    conjugate transpose of the first, and we cannot just do ``np.conj(kraus).T`` because ``kraus`` has 3 indices and we
    just need to transpose the last two.


    Example:
        Here we show that the Choi operator of the identity channel is proportional to
        a maximally entangled Bell :math:`\ket{\phi^+}` state:

    >>> prog = sf.Program(num_subsystems=1)
    >>> C = extract_channel(prog, cutoff_dim=2, representation='choi')
    >>> print(abs(C).reshape((4,4)))
    [[1. 0. 0. 1.]
     [0. 0. 0. 0.]
     [0. 0. 0. 0.]
     [1. 0. 0. 1.]]

    Args:
        prog (Program): program containing the circuit
        cutoff_dim (int): dimension of each index
        representation (str): choice between ``'choi'``, ``'liouville'`` or ``'kraus'``
        vectorize_modes (bool): if True, reshapes the result into rank-4 tensor,
            otherwise it returns a rank-4N tensor, where N is the number of modes

    Returns:
        array: channel, according to the specified options

    Raises:
        TypeError: if the gates used to construct the circuit are not all unitary or channels
    """
    if not is_channel(prog):
        raise TypeError(
            "The circuit definition contains elements that are neither of type Gate nor of type Channel"
        )

    N = prog.init_num_subsystems
    p = _program_in_CJ_rep(prog, cutoff_dim)

    eng = LocalEngine("fock", backend_options={"cutoff_dim": cutoff_dim, "pure": True})
    choi = eng.run(p).state.dm()
    choi = np.einsum("abcd->cdab", _vectorize(choi))

    if representation.lower() == "choi":
        result = choi
        if not vectorize_modes:
            result = _unvectorize(result, N)

    elif representation.lower() == "liouville":
        result = np.einsum("abcd -> dbca", choi)
        if not vectorize_modes:
            result = _unvectorize(result, N)

    elif representation.lower() == "kraus":
        # The liouville operator is the sum of a bipartite product of kraus matrices, so if we vectorize them we obtain
        # a matrix whose eigenvectors are proportional to the vectorized kraus operators
        vectorized_liouville = np.einsum("abcd -> cadb", choi).reshape(
            [cutoff_dim ** (2 * N), cutoff_dim ** (2 * N)]
        )
        eigvals, eigvecs = np.linalg.eig(vectorized_liouville)

        # We keep only those eigenvectors that correspond to non-zero eigenvalues
        eigvecs = eigvecs[:, ~np.isclose(abs(eigvals), 0)]
        eigvals = eigvals[~np.isclose(abs(eigvals), 0)]

        # We rescale the eigenvectors with the sqrt of the eigenvalues (the other sqrt would rescale the right eigenvectors)
        rescaled_eigenvectors = np.einsum("b,ab->ab", np.sqrt(eigvals), eigvecs)

        # Finally we reshape the eigenvectors to form matrices, i.e., the Kraus operators and we make the first index
        # be the one that indexes the list of Kraus operators.
        result = np.einsum(
            "abc->cab", rescaled_eigenvectors.reshape([cutoff_dim ** N, cutoff_dim ** N, -1])
        )

        if not vectorize_modes:
            result = np.einsum(
                np.reshape(result, [-1] + [cutoff_dim] * (2 * N)),
                range(1 + 2 * N),
                [0] + [2 * n + 1 for n in range(N)] + [2 * n + 2 for n in range(N)],
            )
    else:
        raise ValueError("representation {} not supported".format(representation))

    return result



#!/usr/bin/env python3
import strawberryfields as sf
from strawberryfields.ops import *

# initialize engine and program objects
eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": 5})
iqp_circuit = sf.Program(4)

with iqp_circuit.context as q:
    # prepare the input squeezed states
    S = Sgate(-1)
    S | q[0]
    S | q[1]
    S | q[2]
    S | q[3]

    # gates diagonal in the x quadrature
    CZgate(0.5719) | (q[0], q[1])
    Vgate(-1.9782) | q[2]
    Zgate(2.0603)  | q[3]
    Zgate(-0.7804) | q[0]
    Zgate(0.8578)  | q[2]
    CZgate(1.321)  | (q[1], q[2])
    Zgate(0.473)   | q[3]
    CZgate(0.9946) | (q[0], q[3])
    Zgate(0.1223)  | q[3]
    Vgate(0.6157)  | q[0]
    Vgate(0.3110)  | q[1]
    Zgate(-1.243)  | q[2]
    # end circuit

# run the engine
results = eng.run(iqp_circuit)



# Copyright 2019-2020 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A standalone command-line interface for configuring Strawberry Fields and connecting
to the Xanadu cloud platform.
"""

import argparse
import sys

from strawberryfields.api import Connection
from strawberryfields.configuration import ConfigurationError, create_config, store_account
from strawberryfields.engine import RemoteEngine
from strawberryfields.io import load


def main():
    """The Xanadu cloud platform command line interface.

    **Example:**

    The following is a simple example on getting the help message of the cloud platform command
    line interface. It details each of the options available.

    .. code-block:: console

        $ sf
        usage: sf [-h] [--ping] {configure,run} ...

        See below for available options and commands for working with the Xanadu cloud platform.

        General Options:
          -h, --help       show this help message and exit
          --ping, -p       Tests the connection to the remote backend.

        Commands:
          {configure,run}
            configure      Configure each detail of the API connection.
            run            Run a blackbird script.
    """
    parser = create_parser()
    args = parser.parse_args()

    if args.ping:
        ping()
    elif hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


def create_parser():
    """Creates a parser to process the commands and arguments passed to the
    command line interface.

    Returns:
        ArgumentParser: an argument parser object that defines the related
        options
    """
    parser = argparse.ArgumentParser(
        description="See below for available options and commands for working with the Xanadu cloud platform."
    )

    # Setting a title for the general options (requires setting a private
    # attribute)
    parser._optionals.title = "General Options"

    # Adding the pinging general option
    parser.add_argument(
        "--ping", "-p", action="store_true", help="Tests the connection to the remote backend."
    )

    # Adding subparsers configure and input
    subparsers = parser.add_subparsers(title="Commands")

    # Adding the configure subparser
    configure_parser = subparsers.add_parser(
        "configure", help="Configure each detail of the API connection."
    )
    configure_parser.set_defaults(func=configure)

    configure_parser.add_argument(
        "--token",
        "-t",
        type=str,
        help="Configure Strawberry Fields with your Xanadu cloud platform API token.",
    )
    configure_parser.add_argument(
        "--local",
        "-l",
        action="store_true",
        help="Create a local configuration file in the current directory.",
    )

    # Adding the input subparser
    run_parser = subparsers.add_parser("run", help="Run a blackbird script.")
    run_parser.add_argument(
        "input", type=str, help="The filename or path to the blackbird script to run."
    )
    run_parser.set_defaults(func=run_blackbird_script)
    run_parser.add_argument(
        "--output",
        "-o",
        help="Path to the output file, where the results of the program will be written (stdout by default).",
    )

    return parser


def configure(args):
    r"""An auxiliary function for configuring the API connection to the Xanadu
    cloud platform.

    Can be used to simply save the authentication token with default
    configuration options. Alternatively, a wizard is provided for full
    configurability.

    See more details regarding Strawberry Fields configuration and available
    configuration options on the :doc:`/code/sf_configuration` page.

    Args:
        args (ArgumentParser): arguments that were specified on the command
            line stored as attributes in an argument parser object
    """
    if args.token:
        kwargs = {"authentication_token": args.token}
    else:
        kwargs = configuration_wizard()

    if args.local:
        store_account(**kwargs, location="local")
    else:
        store_account(**kwargs)


def ping():
    """Tests the connection to the remote backend."""
    if Connection().ping():
        sys.stdout.write("You have successfully authenticated to the platform!\n")
    else:
        sys.stdout.write("There was a problem when authenticating to the platform!\n")


def configuration_wizard():
    r"""Provides an interactive selection wizard on the command line to
    configure every option for the API connection.

    Default configuration options are provided as defaults to the user.
    Configuration options are detailed in :doc:`/code/sf_configuration`.

    Returns:
        dict[str, Union[str, bool, int]]: the configuration options
    """
    default_config = create_config()["api"]

    # Getting default values that can be used for as messages when getting inputs
    hostname_default = default_config["hostname"]
    ssl_default = "y" if default_config["use_ssl"] else "n"
    port_default = default_config["port"]

    authentication_token = input(
        "Please enter the authentication token to use when connecting: [] "
    )

    if not authentication_token:
        sys.stdout.write("No authentication token was provided, please configure again.")
        sys.exit()

    hostname = (
        input(
            "Please enter the hostname of the server to connect to: [{}] ".format(hostname_default)
        )
        or hostname_default
    )

    ssl_input = (
        input("Should the client attempt to connect over SSL? [{}] ".format(ssl_default))
        or ssl_default
    )
    use_ssl = ssl_input.upper() == "Y"

    port = (
        input("Please enter the port number to connect with: [{}] ".format(port_default))
        or port_default
    )

    kwargs = {
        "authentication_token": authentication_token,
        "hostname": hostname,
        "use_ssl": use_ssl,
        "port": port,
    }
    return kwargs


def run_blackbird_script(args):
    """Run a blackbird script.

    Related arguments:
    * input: the input blackbird script to be run
    * output: the output file to store the results in (optional)

    Args:
        args (ArgumentParser): arguments that were specified on the command
            line stored as attributes in an argument parser object
    """
    try:
        program = load(args.input)
    except FileNotFoundError:
        sys.stdout.write("The {} blackbird script was not found.".format(args.input))
        sys.exit()

    eng = RemoteEngine(program.target)

    sys.stdout.write("Executing program on remote hardware...\n")
    result = eng.run(program)

    if result and result.samples is not None:
        write_script_results(result.samples, output_file=args.output)
    else:
        sys.stdout.write(
            "Ooops! Something went wrong with obtaining the results. Please check the Blackbird script specified and the connection to the remote engine."
        )
        sys.exit()


def write_script_results(samples, output_file=None):
    """Write the results of the script either to a file or to the standard output.

    Args:
        samples (array[float]): array of samples
        output_file (str or None): the path to the output file, None if no output was defined
    """
    if output_file:
        with open(output_file, "w") as file:
            file.write(str(samples))
    else:
        sys.stdout.write(str(samples))



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Tools to construct graph kernels from GBS.

The graph kernel is built by mapping GBS samples from each graph to a feature vector. Similarity
between graphs can then be determined by calculating the overlap between these vectors.

The functionality here is based upon the research papers:
:cite:`bradler2018graph,schuld2019quantum,bradler2019duality`.

.. seealso::

    :ref:`apps-sim-tutorial`

Coarse-graining GBS samples
---------------------------

GBS feature vectors can be composed of probabilities of coarse-grained combinations of elementary
samples. We consider two coarse grainings:

- **Orbits:** Combine all samples that can be made identical under permutation. Orbits are
  written simply as a sorting of integer photon number samples in non-increasing order with the
  zeros at the end removed. For example, ``[1, 1, 2, 0]`` and ``[2, 1, 0, 1]`` both belong to the
  ``[2, 1, 1]`` orbit.

- **Events:** Combine all :math:`k`-photon orbits where the maximum photon count in any mode does
  not exceed a fixed value :math:`n_{\max}` into an event :math:`E_{k, n_{\max}}`. For example,
  orbits ``[2, 1]``, ``[1, 1, 1]`` are part of the :math:`E_{k=3, n_{\max}=2}` event, while
  orbit ``[3]`` is not.

This module provides the following tools for dealing with coarse-grained orbits and events.

Creating a feature vector
-------------------------

A feature vector of a graph can be created by choosing a collection of orbits or events and
evaluating their probabilities with respect to GBS with the embedded graph. These
probabilities are then selected to be elements of the feature vector. Evaluating the
probabilities of orbits or events can be achieved through two approaches:

- **Direct sampling:** infer the probability of orbits or events from a set of sample data.

- **Monte Carlo approximation:** generate samples within a given orbit or event and use them
  to approximate the probability.

In the direct sampling approach, :math:`N` samples are taken from GBS with the embedded graph.
The number that fall within a given orbit or event :math:`E` are counted, resulting in the count
:math:`c_{E}`. The probability :math:`p(E)` of an orbit or event is then approximated as:

.. math::
    p(E) \approx \frac{c_{E}}{N}.

To perform a Monte Carlo estimation of the probability of an orbit or event :math:`E`,
several samples from :math:`E` are drawn uniformly at random using :func:`orbit_to_sample` or
:func:`event_to_sample`. Suppose :math:`N` samples :math:`\{S_{1}, S_{2}, \ldots , S_{N}\}` are
generated. For each sample, this function calculates the probability :math:`p(S_i)` of observing
that sample from a GBS device programmed according to the input graph and mean photon number. The
sum of the probabilities is then rescaled according to the cardinality :math:`|E|` and the total
number of samples:

.. math::
    p(E) \approx \frac{1}{N}\sum_{i=1}^N p(S_i) |E|.

The sample mean of this sum is an estimate of the rescaled probability :math:`p(E)`.

This module provides functions to evaluate the probability of orbits and events through Monte Carlo
approximation.

Similar to the :func:`~.apps.sample.sample` function, the MC estimators include a ``loss`` argument
to specify the proportion of photons lost in the simulated GBS device.

One canonical method of constructing a feature vector is to pick event probabilities
:math:`p_{k} := p_{E_{k, n_{\max}}}` with all events :math:`E_{k, n_{\max}}` having a maximum
photon count :math:`n_{\max}` in each mode. If :math:`\mathbf{k}` is the vector of selected events,
the resultant feature vector is

.. math::
    f_{\mathbf{k}} = (p_{k_{1}}, p_{k_{2}}, \ldots).

This module allows for such feature vectors to be calculated using both the direct sampling and
Monte Carlo approximation methods.
"""
from collections import Counter
from typing import Generator, Union

import networkx as nx
import numpy as np
from scipy.special import factorial

import strawberryfields as sf


def sample_to_orbit(sample: list) -> list:
    """Provides the orbit corresponding to a given sample.

    **Example usage:**

    >>> sample = [1, 2, 0, 0, 1, 1, 0, 3]
    >>> sample_to_orbit(sample)
    [3, 2, 1, 1, 1]

    Args:
        sample (list[int]): a sample from GBS

    Returns:
        list[int]: the orbit of the sample
    """
    return sorted(filter(None, sample), reverse=True)


def sample_to_event(sample: list, max_count_per_mode: int) -> Union[int, None]:
    r"""Provides the event corresponding to a given sample.

    For an input ``max_count_per_mode``, events are expressed here simply by the total photon
    number :math:`k`.

    **Example usage:**

    >>> sample = [1, 2, 0, 0, 1, 1, 0, 3]
    >>> sample_to_event(sample, 4)
    8

    >>> sample_to_event(sample, 2)
    None

    Args:
        sample (list[int]): a sample from GBS
        max_count_per_mode (int): the maximum number of photons counted in any given mode for a
            sample to be categorized as an event. Samples with counts exceeding this value are
            attributed the event ``None``.

    Returns:
        int or None: the event of the sample
    """
    if max(sample) <= max_count_per_mode:
        return sum(sample)

    return None


def orbit_to_sample(orbit: list, modes: int) -> list:
    """Generates a sample selected uniformly at random from the specified orbit.

    **Example usage:**

    >>> orbit_to_sample([2, 1, 1], 6)
    [0, 1, 2, 0, 1, 0]

    Args:
        orbit (list[int]): orbit to generate a sample from
        modes (int): number of modes in the sample

    Returns:
        list[int]: a sample in the orbit
    """
    if modes < len(orbit):
        raise ValueError("Number of modes cannot be smaller than length of orbit")

    sample = orbit + [0] * (modes - len(orbit))
    np.random.shuffle(sample)
    return sample


def event_to_sample(photon_number: int, max_count_per_mode: int, modes: int) -> list:
    """Generates a sample selected uniformly at random from the specified event.

    **Example usage:**

    >>> event_to_sample(4, 2, 6)
    [0, 1, 0, 0, 2, 1]

    Args:
        photon_number (int): number of photons in the event
        max_count_per_mode (int): maximum number of photons per mode in the event
        modes (int): number of modes in the sample

    Returns:
        list[int]: a sample in the event
    """
    if max_count_per_mode < 0:
        raise ValueError("Maximum number of photons per mode must be non-negative")

    if max_count_per_mode * modes < photon_number:
        raise ValueError(
            "No valid samples can be generated. Consider increasing the "
            "max_count_per_mode or reducing the number of photons."
        )

    cards = []
    orbs = []

    for orb in orbits(photon_number):
        if max(orb) <= max_count_per_mode:
            cards.append(orbit_cardinality(orb, modes))
            orbs.append(orb)

    norm = sum(cards)
    prob = [c / norm for c in cards]

    orbit = orbs[np.random.choice(len(prob), p=prob)]

    return orbit_to_sample(orbit, modes)


def orbits(photon_number: int) -> Generator[list, None, None]:
    """Generate all the possible orbits for a given photon number.

    Provides a generator over the integer partitions of ``photon_number``.
    Code derived from `website <http://jeromekelleher.net/generating-integer-partitions.html>`__
    of Jerome Kelleher's, which is based upon an algorithm from Ref. :cite:`kelleher2009generating`.

    **Example usage:**

    >>> o = orbits(5)
    >>> list(o)
    [[1, 1, 1, 1, 1], [2, 1, 1, 1], [3, 1, 1], [2, 2, 1], [4, 1], [3, 2], [5]]

    Args:
        photon_number (int): number of photons to generate orbits from

    Returns:
        Generator[list[int]]: orbits with total photon number adding up to ``photon_number``
    """
    a = [0] * (photon_number + 1)
    k = 1
    y = photon_number - 1
    while k != 0:
        x = a[k - 1] + 1
        k -= 1
        while 2 * x <= y:
            a[k] = x
            y -= x
            k += 1
        l = k + 1
        while x <= y:
            a[k] = x
            a[l] = y
            yield sorted(a[: k + 2], reverse=True)
            x += 1
            y -= 1
        a[k] = x + y
        y = x + y - 1
        yield sorted(a[: k + 1], reverse=True)


def orbit_cardinality(orbit: list, modes: int) -> int:
    """Gives the number of samples belonging to the input orbit.

    For example, there are three possible samples in the orbit [2, 1, 1] with three modes: [1, 1,
    2], [1, 2, 1], and [2, 1, 1]. With four modes, there are 12 samples in total.

    **Example usage:**

    >>> orbit_cardinality([2, 1, 1], 4)
    12

    Args:
        orbit (list[int]): orbit; we count how many samples are contained in it
        modes (int): number of modes in the samples

    Returns:
        int: number of samples in the orbit
    """
    sample = orbit + [0] * (modes - len(orbit))
    counts = list(Counter(sample).values())
    return int(factorial(modes) / np.prod(factorial(counts)))


def event_cardinality(photon_number: int, max_count_per_mode: int, modes: int) -> int:
    r"""Gives the number of samples belonging to the input event.

    For example, for three modes, there are six samples in an :math:`E_{k=2, n_{\max}=2}` event:
    [1, 1, 0], [1, 0, 1], [0, 1, 1], [2, 0, 0], [0, 2, 0], and [0, 0, 2].

    **Example usage:**

    >>> event_cardinality(2, 2, 3)
    6

    Args:
        photon_number (int): number of photons in the event
        max_count_per_mode (int): maximum number of photons per mode in the event
        modes (int): number of modes in counted samples

    Returns:
        int: number of samples in the event
    """
    cardinality = 0

    for orb in orbits(photon_number):
        if max(orb) <= max_count_per_mode:
            cardinality += orbit_cardinality(orb, modes)

    return cardinality


def prob_orbit_mc(
    graph: nx.Graph, orbit: list, n_mean: float = 5, samples: int = 1000, loss: float = 0.0
) -> float:
    r"""Gives a Monte Carlo estimate of the GBS probability of a given orbit according to the input
    graph.

    To make this estimate, several samples from the orbit are drawn uniformly at random using
    :func:`orbit_to_sample`. The GBS probabilities of these samples are then calculated and the
    sum is used to create an estimate of the orbit probability.

    **Example usage:**

    >>> graph = nx.complete_graph(8)
    >>> prob_orbit_mc(graph, [2, 1, 1])
    0.03744

    Args:
        graph (nx.Graph): input graph encoded in the GBS device
        orbit (list[int]): orbit for which to estimate the probability
        n_mean (float): total mean photon number of the GBS device
        samples (int): number of samples used in the Monte Carlo estimation
        loss (float): fraction of photons lost in GBS

    Returns:
        float: estimated orbit probability
    """
    if samples < 1:
        raise ValueError("Number of samples must be at least one")
    if n_mean < 0:
        raise ValueError("Mean photon number must be non-negative")
    if not 0 <= loss <= 1:
        raise ValueError("Loss parameter must take a value between zero and one")

    modes = graph.order()
    photons = sum(orbit)
    A = nx.to_numpy_array(graph)
    mean_photon_per_mode = n_mean / float(modes)

    p = sf.Program(modes)

    # pylint: disable=expression-not-assigned
    with p.context as q:
        sf.ops.GraphEmbed(A, mean_photon_per_mode=mean_photon_per_mode) | q

        if loss:
            for _q in q:
                sf.ops.LossChannel(1 - loss) | _q

    eng = sf.LocalEngine(backend="gaussian")
    result = eng.run(p)

    prob = 0

    for _ in range(samples):
        sample = orbit_to_sample(orbit, modes)
        prob += result.state.fock_prob(sample, cutoff=photons + 1)

    prob = prob * orbit_cardinality(orbit, modes) / samples

    return prob


def prob_event_mc(
    graph: nx.Graph,
    photon_number: int,
    max_count_per_mode: int,
    n_mean: float = 5,
    samples: int = 1000,
    loss: float = 0.0,
) -> float:
    r"""Gives a Monte Carlo estimate of the GBS probability of a given event according to the input
    graph.

    To make this estimate, several samples from the event are drawn uniformly at random using
    :func:`event_to_sample`. The GBS probabilities of these samples are then calculated and the
    sum is used to create an estimate of the event probability.

    **Example usage:**

    >>> graph = nx.complete_graph(8)
    >>> prob_event_mc(graph, 4, 2)
    0.1395

    Args:
        graph (nx.Graph): input graph encoded in the GBS device
        photon_number (int): number of photons in the event
        max_count_per_mode (int): maximum number of photons per mode in the event
        n_mean (float): total mean photon number of the GBS device
        samples (int): number of samples used in the Monte Carlo estimation
        loss (float): fraction of photons lost in GBS

    Returns:
        float: estimated orbit probability
    """
    if samples < 1:
        raise ValueError("Number of samples must be at least one")
    if n_mean < 0:
        raise ValueError("Mean photon number must be non-negative")
    if not 0 <= loss <= 1:
        raise ValueError("Loss parameter must take a value between zero and one")
    if photon_number < 0:
        raise ValueError("Photon number must not be below zero")
    if max_count_per_mode < 0:
        raise ValueError("Maximum number of photons per mode must be non-negative")

    modes = graph.order()
    A = nx.to_numpy_array(graph)
    mean_photon_per_mode = n_mean / float(modes)

    p = sf.Program(modes)

    # pylint: disable=expression-not-assigned
    with p.context as q:
        sf.ops.GraphEmbed(A, mean_photon_per_mode=mean_photon_per_mode) | q

        if loss:
            for _q in q:
                sf.ops.LossChannel(1 - loss) | _q

    eng = sf.LocalEngine(backend="gaussian")
    result = eng.run(p)

    prob = 0

    for _ in range(samples):
        sample = event_to_sample(photon_number, max_count_per_mode, modes)
        prob += result.state.fock_prob(sample, cutoff=photon_number + 1)

    prob = prob * event_cardinality(photon_number, max_count_per_mode, modes) / samples

    return prob


def feature_vector_sampling(
    samples: list, event_photon_numbers: list, max_count_per_mode: int = 2
) -> list:
    r"""Calculates feature vector with respect to input samples.

    The feature vector is composed of event probabilities with a fixed maximum photon count in
    each mode but a range of total photon numbers specified by ``event_photon_numbers``.

    Probabilities are reconstructed by measuring the occurrence of events in the input ``samples``.

    **Example usage:**

    >>> from strawberryfields.apps import data
    >>> samples = data.Mutag0()
    >>> feature_vector_sampling(samples, [2, 4, 6])
    [0.19035, 0.2047, 0.1539]

    Args:
        samples (list[list[int]]): a list of samples
        event_photon_numbers (list[int]): a list of events described by their total photon number
        max_count_per_mode (int): maximum number of photons per mode for all events

    Returns:
        list[float]: a feature vector of event probabilities in the same order as
        ``event_photon_numbers``
    """
    if min(event_photon_numbers) < 0:
        raise ValueError("Cannot request events with photon number below zero")

    if max_count_per_mode < 0:
        raise ValueError("Maximum number of photons per mode must be non-negative")

    n_samples = len(samples)

    e = (sample_to_event(s, max_count_per_mode) for s in samples)
    count = Counter(e)

    return [count[p] / n_samples for p in event_photon_numbers]


def feature_vector_mc(
    graph: nx.Graph,
    event_photon_numbers: list,
    max_count_per_mode: int = 2,
    n_mean: float = 5,
    samples: int = 1000,
    loss: float = 0.0,
) -> list:
    r"""Calculates feature vector using Monte Carlo estimation of event probabilities according to the
    input graph.

    The feature vector is composed of event probabilities with a fixed maximum photon count in
    each mode but a range of total photon numbers specified by ``event_photon_numbers``.

    Probabilities are reconstructed using Monte Carlo estimation.

    **Example usage:**

    >>> graph = nx.complete_graph(8)
    >>> feature_vector_mc(graph, [2, 4, 6], 2)
    [0.2115, 0.1457, 0.09085]

    Args:
        graph (nx.Graph): input graph
        event_photon_numbers (list[int]): a list of events described by their total photon number
        max_count_per_mode (int): maximum number of photons per mode for all events
        n_mean (float): total mean photon number of the GBS device
        samples (int): number of samples used in the Monte Carlo estimation
        loss (float): fraction of photons lost in GBS

    Returns:
        list[float]: a feature vector of event probabilities in the same order as
        ``event_photon_numbers``
    """
    if samples < 1:
        raise ValueError("Number of samples must be at least one")
    if n_mean < 0:
        raise ValueError("Mean photon number must be non-negative")
    if not 0 <= loss <= 1:
        raise ValueError("Loss parameter must take a value between zero and one")
    if min(event_photon_numbers) < 0:
        raise ValueError("Cannot request events with photon number below zero")
    if max_count_per_mode < 0:
        raise ValueError("Maximum number of photons per mode must be non-negative")

    return [
        prob_event_mc(graph, photons, max_count_per_mode, n_mean, samples, loss)
        for photons in event_photon_numbers
    ]



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tensorflow backend interface
============================

"""
# pylint: disable=too-many-public-methods
import numpy as np
import tensorflow as tf

from strawberryfields.backends import BaseFock, ModeMap
from .circuit import Circuit
from .ops import _check_for_eval, mixed, partial_trace, reorder_modes
from .states import FockStateTF


class TFBackend(BaseFock):
    r"""Implements a simulation of quantum optical circuits in a truncated
    Fock basis using `TensorFlow <http://www.numpy.org/>`_, returning a :class:`~.FockStateTF`
    state object.
    """

    short_name = "tf"
    circuit_spec = "tf"

    def __init__(self, graph=None):
        """Initialize a TFBackend object.

        Args:
            graph (tf.Graph): optional Tensorflow Graph object where circuit should be defined
        """
        super().__init__()
        self._supported["mixed_states"] = True
        self._supported["batched"] = True
        self._supported["symbolic"] = True
        if graph is None:
            self._graph = tf.get_default_graph()
        else:
            self._graph = graph
        self._init_modes = None  #: int: initial number of modes in the circuit
        self._modemap = None  #: Modemap: maps external mode indices to internal ones
        self.circuit = (
            None  #: ~.tfbackend.circuit.Circuit: representation of the simulated quantum state
        )

    def _remap_modes(self, modes):
        if isinstance(modes, int):
            modes = [modes]
            was_int = True
        else:
            was_int = False
        map_ = self._modemap.show()
        submap = [map_[m] for m in modes]
        if not self._modemap.valid(modes) or None in submap:
            raise ValueError("The specified modes are not valid.")
        remapped_modes = self._modemap.remap(modes)

        if was_int:
            remapped_modes = remapped_modes[0]
        return remapped_modes

    def begin_circuit(self, num_subsystems, **kwargs):
        r"""Instantiate a quantum circuit.

        Instantiates a representation of a quantum optical state with ``num_subsystems`` modes.
        The state is initialized to vacuum.

        The modes in the circuit are indexed sequentially using integers, starting from zero.
        Once an index is assigned to a mode, it can never be re-assigned to another mode.
        If the mode is deleted its index becomes invalid.
        An operation acting on an invalid or unassigned mode index raises an ``IndexError`` exception.

        Args:
            num_subsystems (int): number of modes in the circuit

        Keyword Args:
            cutoff_dim (int): Numerical Hilbert space cutoff dimension for the modes.
                For each mode, the simulator can represent the Fock states :math:`\ket{0}, \ket{1}, \ldots, \ket{\text{cutoff_dim}-1}`.
            pure (bool): If True (default), use a pure state representation (otherwise will use a mixed state representation).
            batch_size (None or int): Size of the batch-axis dimension. If None, no batch-axis will be used.
        """
        cutoff_dim = kwargs.get("cutoff_dim", None)
        pure = kwargs.get("pure", True)
        batch_size = kwargs.get("batch_size", None)

        if cutoff_dim is None:
            raise ValueError("Argument 'cutoff_dim' must be passed to the Tensorflow backend")

        if not isinstance(num_subsystems, int):
            raise ValueError("Argument 'num_subsystems' must be a positive integer")
        if not isinstance(cutoff_dim, int):
            raise ValueError("Argument 'cutoff_dim' must be a positive integer")
        if not isinstance(pure, bool):
            raise ValueError("Argument 'pure' must be either True or False")
        if batch_size == 1:
            raise ValueError(
                "batch_size of 1 not supported, please use different batch_size or set batch_size=None"
            )

        with tf.name_scope("Begin_circuit"):
            self._modemap = ModeMap(num_subsystems)
            circuit = Circuit(self._graph, num_subsystems, cutoff_dim, pure, batch_size)

        self._init_modes = num_subsystems
        self.circuit = circuit

    def reset(self, pure=True, **kwargs):
        """Reset the circuit so that all the modes are in the vacuum state.

        After the reset the circuit is in the same state as it was after
        the last :meth:`begin_circuit` call. It will have the original number
        of modes, all initialized in the vacuum state. Some circuit parameters
        may be changed during the reset, see the keyword args below.

        Args:
            pure (bool): if True, initialize the circuit in a pure state representation
                (will use a mixed state representation if pure is False)

        Keyword Args:
            cutoff_dim (int): new Hilbert space truncation dimension
            hard (bool): Whether to reset the underlying TensorFlow graph.
                If True (default), then resets the underlying tensor graph as well.
                If False, then the circuit is reset to its initial state, but ops that
                have already been declared are still accessible.
        """
        hard = kwargs.pop("hard", True)
        if hard:
            tf.reset_default_graph()
            self._graph = tf.get_default_graph()

        with tf.name_scope("Reset"):
            self._modemap.reset()
            self.circuit.reset(
                graph=self._graph, num_subsystems=self._init_modes, pure=pure, **kwargs
            )

    def get_cutoff_dim(self):
        return self.circuit.cutoff_dim

    def get_modes(self):
        # pylint: disable=protected-access
        return [i for i, j in enumerate(self._modemap._map) if j is not None]

    def prepare_vacuum_state(self, mode):
        with tf.name_scope("Prepare_vacuum"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_vacuum_state(remapped_mode)

    def prepare_coherent_state(self, alpha, mode):
        with tf.name_scope("Prepare_coherent"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_coherent_state(alpha, remapped_mode)

    def prepare_squeezed_state(self, r, phi, mode):
        with tf.name_scope("Prepare_squeezed"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_squeezed_state(r, phi, remapped_mode)

    def prepare_displaced_squeezed_state(self, alpha, r, phi, mode):
        with tf.name_scope("Prepare_displaced_squeezed"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_displaced_squeezed_state(alpha, r, phi, remapped_mode)

    def prepare_fock_state(self, n, mode):
        with tf.name_scope("Prepare_fock"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_fock_state(n, remapped_mode)

    def prepare_ket_state(self, state, modes):
        with tf.name_scope("Prepare_state"):
            self.circuit.prepare_multimode(state, self._remap_modes(modes), True)

    def prepare_dm_state(self, state, modes):
        with tf.name_scope("Prepare_state"):
            self.circuit.prepare_multimode(state, self._remap_modes(modes), False)

    def prepare_thermal_state(self, nbar, mode):
        with tf.name_scope("Prepare_thermal"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.prepare_thermal_state(nbar, remapped_mode)

    def rotation(self, phi, mode):
        with tf.name_scope("Rotation"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.phase_shift(phi, remapped_mode)

    def displacement(self, alpha, mode):
        with tf.name_scope("Displacement"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.displacement(alpha, remapped_mode)

    def squeeze(self, z, mode):
        with tf.name_scope("Squeeze"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.squeeze(z, remapped_mode)

    def beamsplitter(self, t, r, mode1, mode2):
        with tf.name_scope("Beamsplitter"):
            if isinstance(t, complex):
                raise ValueError("Beamsplitter transmittivity t must be a float.")
            if isinstance(t, tf.Tensor):
                if t.dtype.is_complex:
                    raise ValueError("Beamsplitter transmittivity t must be a float.")
            remapped_modes = self._remap_modes([mode1, mode2])
            self.circuit.beamsplitter(t, r, remapped_modes[0], remapped_modes[1])

    def loss(self, T, mode):
        with tf.name_scope("Loss"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.loss(T, remapped_mode)

    def cubic_phase(self, gamma, mode):
        with tf.name_scope("Cubic_phase"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.cubic_phase(gamma, remapped_mode)

    def kerr_interaction(self, kappa, mode):
        with tf.name_scope("Kerr_interaction"):
            remapped_mode = self._remap_modes(mode)
            self.circuit.kerr_interaction(kappa, remapped_mode)

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        with tf.name_scope("Cross-Kerr_interaction"):
            remapped_modes = self._remap_modes([mode1, mode2])
            self.circuit.cross_kerr_interaction(kappa, remapped_modes[0], remapped_modes[1])

    def state(self, modes=None, **kwargs):
        r"""Returns the state of the quantum simulation, restricted to the subsystems defined by `modes`.

        See :meth:`.BaseBackend.state`.

        Keyword Args:
            session (tf.Session): TensorFlow session
            feed_dict (Dict): Dictionary containing the desired numerical values for Tensors
                for numerically evaluating the state. Used with ``session``.

        Returns:
            FockStateTF: state description
        """
        with tf.name_scope("State"):
            s = self.circuit.state
            pure = self.circuit.state_is_pure
            num_modes = self.circuit.num_modes
            batched = self.circuit.batched

            # reduce rho down to specified subsystems
            if modes is None:
                # reduced state is full state
                reduced_state = s
                modes = list(range(num_modes))
            else:
                if isinstance(modes, int):
                    modes = [modes]
                if len(modes) != len(set(modes)):
                    raise ValueError("The specified modes cannot be duplicated.")
                if len(modes) > num_modes:
                    raise ValueError(
                        "The number of specified modes cannot be larger than the number of subsystems."
                    )

                if pure:
                    # convert to mixed state representation
                    reduced_state = mixed(s, batched)
                    pure = False
                else:
                    reduced_state = s

                    # trace our all modes not in modes
                    # todo: Doing this one by one is very inefficient. The partial trace function should be improved.
                for mode in sorted([m for m in range(num_modes) if m not in modes], reverse=True):
                    reduced_state = partial_trace(reduced_state, mode, False, batched)
                reduced_state_pure = False

            # unless the modes were requested in order, we need to swap indices around
            if modes != sorted(modes):
                mode_permutation = np.argsort(np.argsort(modes))
                reduced_state = reorder_modes(
                    reduced_state, mode_permutation, reduced_state_pure, batched
                )

            evaluate_results, session, feed_dict, close_session = _check_for_eval(kwargs)
            if evaluate_results:
                s = session.run(reduced_state, feed_dict=feed_dict)
                if close_session:
                    session.close()
            else:
                s = reduced_state

            modenames = ["q[{}]".format(i) for i in np.array(self.get_modes())[modes]]
            state_ = FockStateTF(
                s,
                len(modes),
                pure,
                self.circuit.cutoff_dim,
                graph=self._graph,
                batched=batched,
                mode_names=modenames,
                eval=evaluate_results,
            )
        return state_

    def measure_fock(self, modes, shots=1, select=None, **kwargs):
        """Measure the given modes in the Fock basis.

        See :meth:`.BaseFock.measure_fock`.

        Keyword Args:
            session (tf.Session): TensorFlow session
            feed_dict (Dict): Dictionary containing the desired numerical values for Tensors
                for numerically evaluating the measurement results. Used with ``session``.

        Returns:
            tuple[int] or tuple[Tensor]: measurement outcomes
        """
        if shots != 1:
            raise NotImplementedError(
                "TF backend currently does not support " "shots != 1 for Fock measurement"
            )
        with tf.name_scope("Measure_fock"):
            remapped_modes = self._remap_modes(modes)
            meas = self.circuit.measure_fock(remapped_modes, select=select, **kwargs)
        return meas

    def measure_homodyne(self, phi, mode, shots=1, select=None, **kwargs):
        """Perform a homodyne measurement on the specified modes.

        See :meth:`.BaseBackend.measure_homodyne`.

        Keyword Args:
            session (tf.Session): TensorFlow session
            feed_dict (Dict): Dictionary containing the desired numerical values for Tensors
                for numerically evaluating the measurement results. Used with ``session``.
            num_bins (int): Number of equally spaced bins for the probability distribution function
                (pdf) simulating the homodyne measurement (default: 100000).
            max (float): The pdf is discretized onto the 1D grid [-max,max] (default: 10).

        Returns:
            float or tf.Tensor: measurement outcome
        """
        if shots != 1:
            raise NotImplementedError(
                "TF backend currently does not support " "shots != 1 for homodyne measurement"
            )
        with tf.name_scope("Measure_homodyne"):
            remapped_mode = self._remap_modes(mode)
            meas = self.circuit.measure_homodyne(phi, remapped_mode, select, **kwargs)
        return meas

    def is_vacuum(self, tol=0.0, **kwargs):
        with tf.name_scope("Is_vacuum"):
            with self.circuit.graph.as_default():
                vac_elem = self.circuit.vacuum_element()
                if "eval" in kwargs and kwargs["eval"] is False:
                    v = vac_elem
                else:
                    sess = tf.Session()
                    v = sess.run(vac_elem)
                    sess.close()

            result = np.abs(v - 1) <= tol
        return result

    def del_mode(self, modes):
        with tf.name_scope("Del_mode"):
            remapped_modes = self._remap_modes(modes)
            if isinstance(remapped_modes, int):
                remapped_modes = [remapped_modes]
            self.circuit.del_mode(remapped_modes)
            self._modemap.delete(modes)

    def add_mode(self, n=1):
        with tf.name_scope("Add_mode"):
            self.circuit.add_mode(n)
            self._modemap.add(n)

    @property
    def graph(self):
        """
        Get the Tensorflow Graph object where the current quantum circuit is defined.

        Returns:
            tf.Graph: the circuit's graph
        """
        return self._graph



#!/usr/bin/env python3
import numpy as np
import tensorflow as tf
import strawberryfields as sf
from strawberryfields.ops import *

# define width and depth of CV quantum neural network
modes = 1
layers = 8
cutoff_dim = 6

# defining desired state (single photon state)
target_state = np.zeros(cutoff_dim)
target_state[1] = 1
target_state = tf.constant(target_state, dtype=tf.complex64)


# define interferometer
def interferometer(theta, phi, rphi, q):
    """Parameterised interferometer acting on N qumodes

    Args:
        theta (list): list of length N(N-1)/2 real parameters
        phi (list): list of length N(N-1)/2 real parameters
        rphi (list): list of length N-1 real parameters
        q (list): list of qumodes the interferometer is to be applied to
    """
    N = len(q)

    if N == 1:
        # the interferometer is a single rotation
        Rgate(rphi[0]) | q[0]
        return

    n = 0  # keep track of free parameters

    # Apply the rectangular beamsplitter array
    # The array depth is N
    for l in range(N):
        for k, (q1, q2) in enumerate(zip(q[:-1], q[1:])):
            # skip even or odd pairs depending on layer
            if (l + k) % 2 != 1:
                BSgate(theta[n], phi[n]) | (q1, q2)
                n += 1

    # apply the final local phase shifts to all modes except the last one
    for i in range(max(1, N - 1)):
        Rgate(rphi[i]) | q[i]
    # Rgate only applied to first N - 1 modes


# define layer
def layer(q):
    """CV quantum neural network layer acting on N modes

    Args:
        q (list): list of qumodes the layer is to be applied to
    """
    N = len(q)
    BS_variable_number = int(N * (N - 1) / 2)
    R_variable_number = max(1, N - 1)

    theta_variables_1 = tf.Variable(tf.random_normal(shape=[BS_variable_number]))
    phi_variables_1 = tf.Variable(tf.random_normal(shape=[BS_variable_number]))
    rphi_variables_1 = tf.Variable(tf.random_normal(shape=[R_variable_number]))

    theta_variables_2 = tf.Variable(tf.random_normal(shape=[BS_variable_number]))
    phi_variables_2 = tf.Variable(tf.random_normal(shape=[BS_variable_number]))
    rphi_variables_2 = tf.Variable(tf.random_normal(shape=[R_variable_number]))

    s_variables = tf.Variable(tf.random_normal(shape=[N], stddev=0.0001))
    d_variables_r = tf.Variable(tf.random_normal(shape=[N], stddev=0.0001))
    d_variables_phi = tf.Variable(tf.random_normal(shape=[N]))
    k_variables = tf.Variable(tf.random_normal(shape=[N], stddev=0.0001))

    # begin layer
    interferometer(theta_variables_1, phi_variables_1, rphi_variables_1, q)

    for i in range(N):
        Sgate(s_variables[i]) | q[i]

    interferometer(theta_variables_2, phi_variables_2, rphi_variables_2, q)

    for i in range(N):
        Dgate(d_variables_r[i], d_variables_phi[i]) | q[i]
        Kgate(k_variables[i]) | q[i]
    # end layer


# initialize engine and program objects
eng = sf.Engine(backend="tf", backend_options={"cutoff_dim": cutoff_dim})
qnn = sf.Program(modes)

with qnn.context as q:
    for _ in range(layers):
        layer(q)

# starting the engine
results = eng.run(qnn, eval=False)
ket = results.state.ket()

# defining cost function
difference = tf.reduce_sum(tf.abs(ket - target_state))
fidelity = tf.abs(tf.reduce_sum(tf.conj(ket) * target_state)) ** 2
cost = difference

# setting up optimizer
optimiser = tf.train.AdamOptimizer()
minimize_cost = optimiser.minimize(cost)

print("Beginning optimization")

sess = tf.Session()
sess.run(tf.global_variables_initializer())

fidelity_before = sess.run(fidelity)

for i in range(5000):
    sess.run(minimize_cost)

fidelity_after = sess.run(fidelity)

print("Fidelity before optimization: " + str(fidelity_before))
print("Fidelity after optimization: " + str(fidelity_after))
print("\nTarget state: " + str(sess.run(target_state)))
print("Output state: " + str(np.round(sess.run(ket), decimals=3)))



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gaussian circuit operations"""
# pylint: disable=duplicate-code,attribute-defined-outside-init
import numpy as np
from . import ops
from ..shared_ops import changebasis


class GaussianModes:
    """ Base class for representing and operating on a collection of
    continuous variable modes in the symplectic basis as encoded in a
    covariance matrix and a mean vector.
    The modes are initialized in the (multimode) vacuum state,
    The state of the modes is manipulated by calling the various methods."""

    # pylint: disable=too-many-public-methods

    def __init__(self, num_subsystems):
        r"""The class is initialized by providing an integer indicating the number of modes
        Unlike the "standard" covariance matrix for the Wigner function that uses symmetric ordering
        as defined in e.g.
        [1] Gaussian quantum information
        Christian Weedbrook, Stefano Pirandola, Raúl García-Patrón, Nicolas J. Cerf, Timothy C. Ralph, Jeffrey H. Shapiro, and Seth Lloyd
        Rev. Mod. Phys. 84, 621 – Published 1 May 2012
        we define covariance matrices in terms of the following two quantities:
        $$
        N_{i,j} =\langle a_i^\dagger a_j \rangle
        M_{i,j} = \langle a_i a_j \rangle
        $$
        Note that the matrix $N$ is hermitian and the matrix M is symmetric.
        The mean displacements are stored as expectation values of the destruction operator $\alpha_i  = \langle a_i \rangle$
        We also provide functions that give the symmetric ordered covariance matrices and the mean displacement for the quadrature
        operators $q = a+a^\dagger$ and $p = i(a^\dagger -a)$. Note that with these conventions $[q,p]=2 i$.
        For vacuum one has $N_{i,j}=M_{i,j}=alpha_i =0$,
        The quantities $N,M,\alpha$ are stored in the variable nmat, mmat, mean respectively
        """
        # Check validity
        if not isinstance(num_subsystems, int):
            raise ValueError("Number of modes must be an integer")

        self.hbar = 2
        self.reset(num_subsystems)

    def add_mode(self, n=1):
        """add mode to the circuit"""
        newnlen = self.nlen + n
        newnmat = np.zeros((newnlen, newnlen), dtype=complex)
        newmmat = np.zeros((newnlen, newnlen), dtype=complex)
        newmean = np.zeros(newnlen, dtype=complex)
        newactive = list(np.arange(newnlen, dtype=int))

        for i in range(self.nlen):
            newmean[i] = self.mean[i]
            newactive[i] = self.active[i]
            for j in range(self.nlen):
                newnmat[i, j] = self.nmat[i, j]
                newmmat[i, j] = self.mmat[i, j]

        self.mean = newmean
        self.nmat = newnmat
        self.mmat = newmmat
        self.active = newactive
        self.nlen = newnlen

    def del_mode(self, modes):
        """ delete mode from the circuit"""
        if isinstance(modes, int):
            modes = [modes]

        for mode in modes:
            if self.active[mode] is None:
                raise ValueError("Cannot delete mode, mode does not exist")

            self.loss(0.0, mode)
            self.active[mode] = None

    def reset(self, num_subsystems=None):
        """Resets the simulation state.

        Args:
            num_subsystems (int, optional): Sets the number of modes in the reset
                circuit. None means unchanged.
        """
        if num_subsystems is not None:
            if not isinstance(num_subsystems, int):
                raise ValueError("Number of modes must be an integer")
            self.nlen = num_subsystems

        self.nmat = np.zeros((self.nlen, self.nlen), dtype=complex)
        self.mmat = np.zeros((self.nlen, self.nlen), dtype=complex)
        self.mean = np.zeros(self.nlen, dtype=complex)
        self.active = list(np.arange(self.nlen, dtype=int))

    def get_modes(self):
        """return the modes currently active"""
        return [x for x in self.active if x is not None]

    def displace(self, beta, i):
        """ Implements a displacement operation by the complex number beta in mode i"""
        # Update displacement of mode i by the complex amount bet
        if self.active[i] is None:
            raise ValueError("Cannot displace mode, mode does not exist")

        self.mean[i] += beta

    def squeeze(self, r, phi, k):
        """ Implements a squeezing operation in mode k by the amount z = r*exp(1j*phi)."""
        if self.active[k] is None:
            raise ValueError("Cannot squeeze mode, mode does not exist")

        phase = np.exp(1j * phi)
        phase2 = phase * phase
        sh = np.sinh(r)
        ch = np.cosh(r)
        sh2 = sh * sh
        ch2 = ch * ch
        shch = sh * ch
        nk = np.copy(self.nmat[k])
        mk = np.copy(self.mmat[k])

        alphak = np.copy(self.mean[k])
        # Update displacement of mode k
        self.mean[k] = alphak * ch - phase * np.conj(alphak) * sh
        # Update covariance matrix elements. Only the k column and row of nmat and mmat need to be updated.
        # First update the diagonal elements
        self.nmat[k, k] = (
            sh2
            - phase * shch * np.conj(mk[k])
            - shch * np.conj(phase) * mk[k]
            + ch2 * nk[k]
            + sh2 * nk[k]
        )
        self.mmat[k, k] = (
            -(phase * shch) + phase2 * sh2 * np.conj(mk[k]) + ch2 * mk[k] - 2 * phase * shch * nk[k]
        )

        # Update the column k
        for l in np.delete(np.arange(self.nlen), k):
            self.nmat[k, l] = -(sh * np.conj(phase) * mk[l]) + ch * nk[l]
            self.mmat[k, l] = ch * mk[l] - phase * sh * nk[l]

        # Update row k
        self.nmat[:, k] = np.conj(self.nmat[k])
        self.mmat[:, k] = self.mmat[k]

    def phase_shift(self, phi, k):
        """ Implements a phase shift in mode k by the amount phi."""
        if self.active[k] is None:
            raise ValueError("Cannot phase shift mode, mode does not exist")

        phase = np.exp(1j * phi)
        phase2 = phase * phase
        # Update displacement of mode k
        self.mean[k] = self.mean[k] * phase

        # Update covariance matrix elements. Only the k column and row of nmat and mmat need to be updated.
        # First update the diagonal elements
        self.mmat[k][k] = phase2 * self.mmat[k][k]

        # Update the column k
        for l in np.delete(np.arange(self.nlen), k):
            self.nmat[k][l] = np.conj(phase) * self.nmat[k][l]
            self.mmat[k][l] = phase * self.mmat[k][l]

        # Update row k
        self.nmat[:, k] = np.conj(self.nmat[k])
        self.mmat[:, k] = self.mmat[k]

    def beamsplitter(self, theta, phi, k, l):
        """ Implements a beam splitter operation between modes k and l by the amount theta, phi"""
        if self.active[k] is None or self.active[l] is None:
            raise ValueError("Cannot perform beamsplitter, mode(s) do not exist")

        if k == l:
            raise ValueError("Cannot use the same mode for beamsplitter inputs")

        phase = np.exp(1j * phi)
        phase2 = phase * phase
        sh = np.sin(theta)
        ch = np.cos(theta)
        sh2 = sh * sh
        ch2 = ch * ch
        shch = sh * ch
        # alpha1 = self.mean[0]

        nk = np.copy(self.nmat[k])
        mk = np.copy(self.mmat[k])
        nl = np.copy(self.nmat[l])
        ml = np.copy(self.mmat[l])
        # Update displacement of mode k and l
        alphak = np.copy(self.mean[k])
        alphal = np.copy(self.mean[l])
        self.mean[k] = ch * alphak + phase * sh * alphal
        self.mean[l] = ch * alphal - np.conj(phase) * sh * alphak
        # Update covariance matrix elements. Only the k and l columns and rows of nmat and mmat need to be updated.
        # First update the (k,k), (k,l), (l,l), and (l,l) elements
        self.nmat[k][k] = (
            ch2 * nk[k] + phase * shch * nk[l] + shch * np.conj(phase) * nl[k] + sh2 * nl[l]
        )
        self.nmat[k][l] = (
            -(shch * np.conj(phase) * nk[k])
            + ch2 * nk[l]
            - sh2 * np.conj(phase2) * nl[k]
            + shch * np.conj(phase) * nl[l]
        )
        self.nmat[l][k] = np.conj(self.nmat[k][l])
        self.nmat[l][l] = (
            sh2 * nk[k] - phase * shch * nk[l] - shch * np.conj(phase) * nl[k] + ch2 * nl[l]
        )

        self.mmat[k][k] = ch2 * mk[k] + 2 * phase * shch * ml[k] + phase2 * sh2 * ml[l]
        self.mmat[k][l] = (
            -(shch * np.conj(phase) * mk[k]) + ch2 * ml[k] - sh2 * ml[k] + phase * shch * ml[l]
        )
        self.mmat[l][k] = self.mmat[k][l]
        self.mmat[l][l] = (
            sh2 * np.conj(phase2) * mk[k] - 2 * shch * np.conj(phase) * ml[k] + ch2 * ml[l]
        )

        # Update columns k and l
        for i in np.delete(np.arange(self.nlen), (k, l)):
            self.nmat[k][i] = ch * nk[i] + sh * np.conj(phase) * nl[i]
            self.mmat[k][i] = ch * mk[i] + phase * sh * ml[i]
            self.nmat[l][i] = -(phase * sh * nk[i]) + ch * nl[i]
            self.mmat[l][i] = -(sh * np.conj(phase) * mk[i]) + ch * ml[i]

        # Update rows k and l
        self.nmat[:, k] = np.conj(self.nmat[k])
        self.mmat[:, k] = self.mmat[k]
        self.nmat[:, l] = np.conj(self.nmat[l])
        self.mmat[:, l] = self.mmat[l]

    def scovmatxp(self):
        r"""Constructs and returns the symmetric ordered covariance matrix in the xp ordering.

        The order for the canonical operators is :math:`q_1,..,q_n, p_1,...,p_n`.
        This differs from the ordering used in [1] which is :math:`q_1,p_1,q_2,p_2,...,q_n,p_n`
        Note that one ordering can be obtained from the other by using a permutation matrix.

        Said permutation matrix is implemented in the function changebasis(n) where n is
        the number of modes.
        """
        mm11 = (
            self.nmat
            + np.transpose(self.nmat)
            + self.mmat
            + np.conj(self.mmat)
            + np.identity(self.nlen)
        )
        mm12 = 1j * (
            -np.transpose(self.mmat)
            + np.transpose(np.conj(self.mmat))
            + np.transpose(self.nmat)
            - self.nmat
        )
        mm22 = (
            self.nmat
            + np.transpose(self.nmat)
            - self.mmat
            - np.conj(self.mmat)
            + np.identity(self.nlen)
        )
        return np.concatenate(
            (
                np.concatenate((mm11, mm12), axis=1),
                np.concatenate((np.transpose(mm12), mm22), axis=1),
            ),
            axis=0,
        ).real

    def smeanxp(self):
        r"""Constructs and returns the symmetric ordered vector of mean in the xp ordering.

        The order for the canonical operators is :math:`q_1, \ldots, q_n, p_1, \ldots, p_n`.
        This differs from the ordering used in [1] which is :math:`q_1, p_1, q_2, p_2, \ldots, q_n, p_n`.
        Note that one ordering can be obtained from the other by using a permutation matrix.

        Said permutation matrix is implemented in the function changebasis(n) where n is
        the number of modes.
        """
        nmodes = self.nlen
        r = np.empty(2 * nmodes)
        r[0:nmodes] = 2 * self.mean.real
        r[nmodes : 2 * nmodes] = 2 * self.mean.imag
        return r

    def scovmat(self):
        """Constructs and returns the symmetric ordered covariance matrix as defined in [1]
        """
        rotmat = changebasis(self.nlen)
        return np.dot(np.dot(rotmat, self.scovmatxp()), np.transpose(rotmat))

    def smean(self):
        r"""the symmetric mean $[q_1,p_1,q_2,p_2,...,q_n,p_n]$"""
        r = np.empty(2 * self.nlen)
        for i in range(self.nlen):
            r[2 * i] = 2 * self.mean[i].real
            r[2 * i + 1] = 2 * self.mean[i].imag
        return r

    def fromsmean(self, r, modes=None):
        r"""Populates the means from a provided vector of means with hbar=2 assumed.

        Args:
            r (array): vector of means in :math:`(x_1,p_1,x_2,p_2,\dots)` ordering
            modes (Sequence): sequence of modes corresponding to the vector of means
        """
        mode_list = modes
        if modes is None:
            mode_list = range(self.nlen)

        for idx, mode in enumerate(mode_list):
            self.mean[mode] = 0.5 * (r[2 * idx] + 1j * r[2 * idx + 1])

    def fromscovmat(self, V, modes=None):
        r"""Updates the circuit's state when a standard covariance matrix is provided.

        Args:
            V (array): covariance matrix in symmetric ordering
            modes (Sequence): sequence of modes corresponding to the covariance matrix
        """
        if modes is None:
            n = len(V) // 2
            modes = np.arange(self.nlen)

            if n != self.nlen:
                raise ValueError(
                    "Covariance matrix is the incorrect size, does not match means vector."
                )
        else:
            n = len(modes)
            modes = np.array(modes)
            if n > self.nlen:
                raise ValueError("Covariance matrix is larger than the number of subsystems.")

        # convert to xp ordering
        rotmat = changebasis(n)
        VV = np.dot(np.dot(np.transpose(rotmat), V), rotmat)

        A = VV[0:n, 0:n]
        B = VV[0:n, n : 2 * n]
        C = VV[n : 2 * n, n : 2 * n]
        Bt = np.transpose(B)

        if n < self.nlen:
            # reset modes to be prepared back to the vacuum state
            for mode in modes:
                self.loss(0.0, mode)

        rows = modes.reshape(-1, 1)
        cols = modes.reshape(1, -1)
        self.nmat[rows, cols] = 0.25 * (A + C + 1j * (B - Bt) - 2 * np.identity(n))
        self.mmat[rows, cols] = 0.25 * (A - C + 1j * (B + Bt))

    def qmat(self, modes=None):
        """ Construct the covariance matrix for the Q function"""
        if modes is None:
            modes = list(range(self.nlen))

        rows = np.reshape(modes, [-1, 1])
        cols = np.reshape(modes, [1, -1])

        sigmaq = np.concatenate(
            (
                np.concatenate(
                    (self.nmat[rows, cols], np.conjugate(self.mmat[rows, cols])), axis=1
                ),
                np.concatenate(
                    (self.mmat[rows, cols], np.conjugate(self.nmat[rows, cols])), axis=1
                ),
            ),
            axis=0,
        ) + np.identity(2 * len(modes))
        return sigmaq

    def fidelity_coherent(self, alpha, modes=None):
        """ Returns a function that evaluates the Q function of the given state """
        if modes is None:
            modes = list(range(self.nlen))

        Q = self.qmat(modes)
        Qi = np.linalg.inv(Q)
        delta = self.mean[modes] - alpha

        delta = np.concatenate((delta, np.conjugate(delta)))
        return np.sqrt(np.linalg.det(Qi).real) * np.exp(
            -0.5 * np.dot(delta, np.dot(Qi, np.conjugate(delta))).real
        )

    def fidelity_vacuum(self, modes=None):
        """fidelity of the current state with the vacuum state"""
        if modes is None:
            modes = list(range(self.nlen))

        alpha = np.zeros(len(modes))
        return self.fidelity_coherent(alpha)

    def Amat(self):
        """ Constructs the A matrix from Hamilton's paper"""
        ######### this needs to be conjugated
        sigmaq = np.concatenate(
            (
                np.concatenate((np.transpose(self.nmat), self.mmat), axis=1),
                np.concatenate((np.transpose(np.conjugate(self.mmat)), self.nmat), axis=1),
            ),
            axis=0,
        ) + np.identity(2 * self.nlen)
        return np.dot(ops.xmat(self.nlen), np.identity(2 * self.nlen) - np.linalg.inv(sigmaq))

    def loss(self, T, k):
        r"""Implements a loss channel in mode k by amplitude loss amount \sqrt{T}
        (energy loss amount T)"""
        if self.active[k] is None:
            raise ValueError("Cannot apply loss channel, mode does not exist")

        sqrtT = np.sqrt(T)
        self.nmat[k] = sqrtT * self.nmat[k]
        self.mmat[k] = sqrtT * self.mmat[k]

        self.nmat[k][k] = sqrtT * self.nmat[k][k]
        self.mmat[k][k] = sqrtT * self.mmat[k][k]

        self.nmat[:, k] = np.conj(self.nmat[k])
        self.mmat[:, k] = self.mmat[k]
        self.mean[k] = sqrtT * self.mean[k]

    def thermal_loss(self, T, nbar, k):
        r""" Implements the thermal loss channel in mode k by amplitude loss amount \sqrt{T}
        unlike the loss channel, here the ancilliary mode that goes into the second arm of the
        beam splitter is prepared in a thermal state with mean photon number nth """
        if self.active[k] is None:
            raise ValueError("Cannot apply loss channel, mode does not exist")

        self.loss(T, k)
        self.nmat += (1 - T) * nbar

    def init_thermal(self, population, mode):
        """ Initializes a state of mode in a thermal state with the given population"""
        self.loss(0.0, mode)
        self.nmat[mode][mode] = population

    def is_vacuum(self, tol=0.0):
        """ Checks if the state is vacuum by calculating its fidelity with vacuum """
        fid = self.fidelity_vacuum()
        return np.abs(fid - 1) <= tol

    def measure_dyne(self, covmat, indices, shots=1):
        """ Performs the general-dyne measurement specified in covmat, the indices should correspond
        with the ordering of the covmat of the measurement
        covmat specifies a gaussian effect via its covariance matrix. For more information see
        Quantum Continuous Variables: A Primer of Theoretical Methods
        by Alessio Serafini page 129
        """
        if covmat.shape != (2 * len(indices), 2 * len(indices)):
            raise ValueError("Covariance matrix size does not match indices provided")

        for i in indices:
            if self.active[i] is None:
                raise ValueError("Cannot apply homodyne measurement, mode does not exist")

        expind = np.concatenate((2 * np.array(indices), 2 * np.array(indices) + 1))
        mp = self.scovmat()
        (A, B, C) = ops.chop_in_blocks(mp, expind)
        V = A - np.dot(np.dot(B, np.linalg.inv(C + covmat)), np.transpose(B))
        V1 = ops.reassemble(V, expind)
        self.fromscovmat(V1)

        r = self.smean()
        (va, vc) = ops.chop_in_blocks_vector(r, expind)
        vm = np.random.multivariate_normal(vc, C, size=shots)
        # The next line is a hack in that it only updates conditioned on the first samples value
        # should still work if shots = 1
        va = va + np.dot(np.dot(B, np.linalg.inv(C + covmat)), vm[0] - vc)
        va = ops.reassemble_vector(va, expind)
        self.fromsmean(va)
        return vm

    def homodyne(self, n, shots=1, eps=0.0002):
        """ Performs a homodyne measurement by calling measure dyne an giving it the
        covariance matrix of a squeezed state whose x quadrature has variance eps**2"""
        covmat = np.diag(np.array([eps ** 2, 1.0 / eps ** 2]))
        res = self.measure_dyne(covmat, [n], shots=shots)

        return res

    def post_select_homodyne(self, n, val, eps=0.0002):
        """ Performs a homodyne measurement but postelecting on the value vals for mode n """
        if self.active[n] is None:
            raise ValueError("Cannot apply homodyne measurement, mode does not exist")
        covmat = np.diag(np.array([eps ** 2, 1.0 / eps ** 2]))
        indices = [n]
        expind = np.concatenate((2 * np.array(indices), 2 * np.array(indices) + 1))
        mp = self.scovmat()
        (A, B, C) = ops.chop_in_blocks(mp, expind)
        V = A - np.dot(np.dot(B, np.linalg.inv(C + covmat)), np.transpose(B))
        V1 = ops.reassemble(V, expind)
        self.fromscovmat(V1)

        r = self.smean()
        (va, vc) = ops.chop_in_blocks_vector(r, expind)
        vm1 = np.random.normal(vc[1], np.sqrt(C[1][1]))
        vm = np.array([val, vm1])
        va = va + np.dot(np.dot(B, np.linalg.inv(C + covmat)), vm - vc)
        va = ops.reassemble_vector(va, expind)
        self.fromsmean(va)
        return val

    def post_select_heterodyne(self, n, alpha_val):
        """ Performs a homodyne measurement but postelecting on the value vals for mode n """
        if self.active[n] is None:
            raise ValueError("Cannot apply heterodyne measurement, mode does not exist")

        covmat = np.identity(2)
        indices = [n]
        expind = np.concatenate((2 * np.array(indices), 2 * np.array(indices) + 1))
        mp = self.scovmat()
        (A, B, C) = ops.chop_in_blocks(mp, expind)
        V = A - np.dot(np.dot(B, np.linalg.inv(C + covmat)), np.transpose(B))
        V1 = ops.reassemble(V, expind)
        self.fromscovmat(V1)

        r = self.smean()
        (va, vc) = ops.chop_in_blocks_vector(r, expind)
        vm = 2.0 * np.array([np.real(alpha_val), np.imag(alpha_val)])
        va = va + np.dot(np.dot(B, np.linalg.inv(C + covmat)), vm - vc)
        va = ops.reassemble_vector(va, expind)
        self.fromsmean(va)
        return alpha_val

    def apply_u(self, U):
        """ Transforms the state according to the linear optical unitary that maps a[i] \to U[i, j]^*a[j]"""
        self.mean = np.dot(np.conj(U), self.mean)
        self.nmat = np.dot(np.dot(U, self.nmat), np.conj(np.transpose(U)))
        self.mmat = np.dot(np.dot(np.conj(U), self.mmat), np.conj(np.transpose(U)))



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fock backend proper"""
# pylint: disable=too-many-arguments,len-as-condition,attribute-defined-outside-init
# pylint: disable=too-many-branches,too-many-locals,too-many-public-methods

import copy
import string
import numbers
from itertools import product

import numpy as np
from numpy import sqrt, pi
from scipy.special import factorial as bang
from numba import jit
from numba.typed import List

from . import ops

def_type = np.complex128
indices = string.ascii_lowercase
MAX_MODES = len(indices) - 3


class Circuit:
    """
    Class implementing a basic simulator for a collection of modes
    in the fock basis.
    """

    def __init__(self, num, trunc, pure=True, do_checks=False):
        r"""Class initializer.

        Args:
            num (non-negative int): Number of modes in the register.
            trunc (positive int): Truncation parameter.  Fock states up to |trunc-1> are representable.
            pure (bool, optional): Whether states are pure (True) or mixed (False)
            do_checks (bool, optional): Whether arguments are to be checked first
            mode (str, optional): Whether to use BLAS or einsum for matrix operations.
        """

        # Check validity
        if num < 0:
            raise ValueError("Number of modes must be non-negative -- got {}".format(num))
        if num > MAX_MODES:
            raise ValueError("Fock simulator has a maximum of {} modes".format(MAX_MODES))
        if trunc <= 0:
            raise ValueError("Truncation must be positive -- got {}".format(trunc))

        self._num_modes = num
        self._hbar = 2
        self._checks = do_checks
        self.reset(pure=pure, cutoff_dim=trunc)

    def _apply_channel(self, kraus_ops, modes):
        """Master channel application function. Applies a channel represented by
        Kraus operators.

        .. note::
                Always results in a mixed state.

        Args:
            kraus_ops (list<array>): A list of Kraus operators
            modes (list<non-negative int>): The modes to apply the channel to
        """

        if self._pure:
            self._state = ops.mix(self._state, self._num_modes)
            self._pure = False

        if len(kraus_ops) == 0:
            self._state = np.zeros(
                [self._trunc for i in range(self._num_modes * 2)], dtype=ops.def_type
            )
        else:
            states = [self.apply_gate_BLAS(k, modes, pure=False) for k in kraus_ops]
            self._state = sum(states)

    def reset(self, pure=None, cutoff_dim=None, num_subsystems=None):
        """Resets the simulation state.

        Args:
            pure (bool, optional): Sets the purity setting. Default is unchanged.
            cutoff_dim (int): New Hilbert space truncation dimension.
            num_subsystems (int, optional): Sets the number of modes in the reset
                circuit. Default is unchanged.
        """
        if pure is not None:
            if not isinstance(pure, bool):
                raise ValueError("Argument 'pure' must be either True or False")
            self._pure = pure

        if num_subsystems is not None:
            if not isinstance(num_subsystems, int):
                raise ValueError("Argument 'num_subsystems' must be a positive integer")
            self._num_modes = num_subsystems

        if cutoff_dim is not None:
            if not isinstance(cutoff_dim, int) or cutoff_dim < 1:
                raise ValueError("Argument 'cutoff_dim' must be a positive integer")
            self._trunc = cutoff_dim

        if self._pure:
            self._state = ops.vacuumState(self._num_modes, self._trunc)
        else:
            self._state = ops.vacuumStateMixed(self._num_modes, self._trunc)

    def apply_gate_BLAS(self, mat, modes, **kwargs):
        """Applies a specified gate to the state.

        Gate application based on custom indexing and matrix multiplication.
        Assumes the input matrix has shape (out1, in1, ...).

        This implementation uses indexing and BLAS. As per stack overflow,
        einsum doesn't actually use BLAS but rather a c implementation. In theory
        if reshaping is efficient this should be faster.

        Args:
            mat (array[complex]): The numeric operator to be applied to the state, of shape ``[trunc]*(2*n)``
            modes (list[int]): The list of modes to which the operator is applied on

        Keyword args:
            state (array[complex]): The state that the operator is applied to
            pure (bool): Whether the state is pure or mixed
            n (int): The total number of modes

        Returns:
            array[complex]: The state after application of the two-mode operation
        """
        # pylint: disable=too-many-branches

        state = kwargs.get("state", self._state)
        pure = kwargs.get("pure", self._pure)
        n = kwargs.get("n", self._num_modes)

        size = len(modes)
        dim = self._trunc ** size
        stshape = [self._trunc for i in range(size)]

        # Apply the following matrix transposition:
        # |m1><m1| |m2><m2| ... |mn><mn| -> |m1>|m2>...|mn><m1|<m2|...<mn|
        transpose_list = [2 * i for i in range(size)] + [2 * i + 1 for i in range(size)]
        matview = np.transpose(mat, transpose_list).reshape((dim, dim))

        # checks if matview is diagonal (for example, Kerr/rotation gates) and, if so, use faster contractions
        diag = np.all(matview == np.diag(np.diagonal(matview)))

        if pure:
            if n == 1:
                if diag:
                    mat_diag = mat.diagonal()
                    return np.multiply(mat_diag, state)
                return np.dot(mat, state)

            # Transpose the state into the following form:
            # |psi> |mode[0]> |mode[1]> ... |mode[n]>
            transpose_list = [i for i in range(n) if not i in modes] + modes
            view = np.transpose(state, transpose_list)

            # Apply matrix to each substate
            ret = np.zeros([self._trunc for i in range(n)], dtype=def_type)
            for i in product(*([range(self._trunc) for j in range(n - size)])):
                if diag:
                    mat_diag = matview.diagonal()
                    ret[i] = np.multiply(mat_diag, view[i].ravel()).reshape(stshape)
                else:
                    ret[i] = np.dot(matview, view[i].ravel()).reshape(stshape)

            # "untranspose" the return matrix ret
            untranspose_list = [0] * len(transpose_list)
            for i in range(len(transpose_list)):  # pylint: disable=consider-using-enumerate
                untranspose_list[transpose_list[i]] = i

            return np.transpose(ret, untranspose_list)

        # otherwise, the state is mixed
        if n == 1:
            if diag:
                mat_diag = mat.diagonal().reshape(-1, 1)
                return np.multiply(mat_diag, np.multiply(state, mat_diag.conj().T))
            return np.dot(mat, np.dot(state, mat.conj().T))

        # Transpose the state into the following form:
        # |psi><psi||mode[0]>|mode[1]>...|mode[n]><mode[0]|<mode[1]|...<mode[n]|
        transpose_list = [i for i in range(n * 2) if not i // 2 in modes]
        transpose_list = transpose_list + [2 * i for i in modes] + [2 * i + 1 for i in modes]
        view = np.transpose(state, transpose_list)

        # Apply matrix to each substate
        ret = np.zeros([self._trunc for i in range(n * 2)], dtype=def_type)
        for i in product(*([range(self._trunc) for j in range((n - size) * 2)])):
            if diag:
                mat_diag = matview.diagonal().reshape(-1, 1)
                ret[i] = np.multiply(
                    mat_diag, np.multiply(view[i].reshape((dim, dim)), mat_diag.conj().T)
                ).reshape(stshape + stshape)
            else:
                ret[i] = np.dot(
                    matview, np.dot(view[i].reshape((dim, dim)), matview.conj().T)
                ).reshape(stshape + stshape)

        # "untranspose" the return matrix ret
        untranspose_list = [0] * len(transpose_list)
        for i in range(len(transpose_list)):  # pylint: disable=consider-using-enumerate
            untranspose_list[transpose_list[i]] = i

        return np.transpose(ret, untranspose_list)

    def apply_twomode_gate(self, mat, modes, gate="BSgate"):
        """Applies a two-mode gate to the state.

        Applies the specified two-mode gate to the state using custom tensor contractions and
        the Numba compiler for faster application. Currently, only the beamsplitter and the
        two-mode squeeze gate are supported.

        Args:
            mat (array[complex]): The numeric operator to be applied to the state, of shape `[trunc]*(2*n)`
            modes (list[int]): The list of modes to which the operator is applied on
            gate (str): The gate that is being applied. This argument determines the selection rules that
                are used. Options are ``"BSgate"`` and ``"S2gate"``.

        Returns:
            array[complex]: The state after application of the two-mode operation
        """
        if self._pure:
            t1 = modes[0]
            t2 = modes[1]

            # moves the modes on which the gate should be applied to the
            # front indices of the state, for mode 1 and 2 respectively
            # e.g. from [i1, i2, i3] --> [i2, i3, i1] if the gate is applied to modes 2 and 3
            switch_list_1 = np.arange(self._num_modes)
            switch_list_2 = np.arange(self._num_modes)
            switch_list_1[[0, t1]] = switch_list_1[[t1, 0]]
            switch_list_2[[1, t2]] = switch_list_2[[t2, 1]]

            self._state = self._state.transpose(switch_list_1)
            self._state = self._state.transpose(switch_list_2)

            if gate == "BSgate":
                self._state = self._apply_BS(mat, self._state, self._trunc)
            elif gate == "S2gate":
                self._state = self._apply_S2(mat, self._state, self._trunc)
            else:
                raise NotImplementedError(
                    "Currently, selection rules are only implemented for the BSgate "
                    "and the S2gate. The {} gate is not supported".format(gate)
                )

            self._state = self._state.transpose(switch_list_2)
            ret = self._state.transpose(switch_list_1)
        else:
            t1 = 2 * modes[0]
            t2 = 2 * modes[1]

            # moves the modes on which the gate should be applied to the
            # front indices of the state, for mode 1 and 2 respectively
            # e.g. from [i1, j1, i2, j2, i3, j3] --> [i2, j2, i3, j3, i1, j1]
            # if the gate is applied to modes 2 and 3
            switch_list_1 = np.arange(2 * self._num_modes)
            switch_list_2 = np.arange(2 * self._num_modes)
            switch_list_1[[0, 1, t1, t1 + 1]] = switch_list_1[[t1, t1 + 1, 0, 1]]
            switch_list_2[[0, 1, t2, t2 + 1]] = switch_list_2[[t2, t2 + 1, 0, 1]]

            # puts the modes on which the gate should be applied together
            # e.g. from [i2, j2, i3, j3, i1, j1] --> [i2, i3, j2, j3, i1, j1]
            # if the gate is applied to modes 2 and 3 (which are already switched to the first two modes)
            transpose_list = np.arange(2 * self._num_modes)
            transpose_list[[t1 + 1, t2]] = transpose_list[[t2, t1 + 1]]

            self._state = self._state.transpose(transpose_list)
            self._state = self._state.transpose(switch_list_1)

            if gate == "BSgate":
                self._state = self._apply_BS(mat, self._state, self._trunc)
                self._state = self._state.transpose(switch_list_1)

                self._state = self._state.transpose(switch_list_2)
                self._state = self._apply_BS(mat.conj(), self._state, self._trunc)

            elif gate == "S2gate":
                self._state = self._apply_S2(mat, self._state, self._trunc)
                self._state = self._state.transpose(switch_list_1)

                self._state = self._state.transpose(switch_list_2)
                self._state = self._apply_S2(mat.conj(), self._state, self._trunc)

            else:
                raise NotImplementedError

            self._state = self._state.transpose(switch_list_2)
            ret = self._state.transpose(transpose_list)

        return ret

    # ignored in covtest (doesn't work well with the jit decorator)
    @staticmethod
    @jit(nopython=True)
    def _apply_BS(mat, state, trunc):  # pragma: no cover
        r"""Applies the BS gate to the first bra in state.

        The beamsplitter matrix elements :math:`B_{ij}^{kl}` satisfy the selection
        rules

        .. math::
            B_{ij}^{kl} \propto \delta_{i+j}^{k+l}

        This implies that one can contract pure states (without loss of generality
        we assume only two modes) using one less loop as follows:

        .. math::
            c'_{i, j} = \sum\limits_{k = \text{max}(1+i+j-n, 0)}
                    ^{\text{min}(i+j, n-1) + 1}
            B_{ik}^{j, i+j-k} c_{k, i+j-k}

        where :math:`c_{kl}` is the tensor representing the two mode state and
        :math:`n` is the cutoff in Fock space. Similarly for a mixed state it is
        possible to do the update of the state using two less summations.
        """
        ret = np.zeros_like(state, dtype=np.complex128)
        for i in range(trunc):
            for j in range(trunc):
                for k in range(max(1 + i + j - trunc, 0), min(i + j, trunc - 1) + 1):
                    ret[i, j] += mat[i, k, j, i + j - k] * state[k, i + j - k]
        return ret

    # ignored in covtest (doesn't work well with the jit decorator)
    @staticmethod
    @jit(nopython=True)
    def _apply_S2(mat, state, trunc):  # pragma: no cover
        r"""Applies the S2 gate to the first bra in state.

        The two-mode squeeze matrix elements :math:`T_{ij}^{kl}` satisfy the selection
        rules

        .. math::
            T_{ij}^{kl} \propto \delta_{i-j}^{k-l}

        This implies that one can contract pure states (without loss of generality
        we assume only two modes) using one less loop as follows:

        .. math::
            c'_{i, k} = \sum\limits_{k = \text{max}(i-j, 0)}^{\text{min}(i-j, 0) + n}
            T_{ij}^{k, k+j-i} c_{j, k+j-i}

        where :math:`c_{jl}` is the tensor representing the two mode state and
        :math:`n` is the cutoff in Fock space. Similarly for a mixed state it is
        possible to do the update of the state using two less summations.
        """
        ret = np.zeros_like(state, dtype=np.complex128)
        for i in range(trunc):
            for j in range(trunc):
                for k in range(max(i - j, 0), trunc + min(i - j, 0)):
                    ret[i, k] += mat[i, j, k, k + j - i] * state[j, k + j - i]
        return ret

    def norm(self):
        """returns the norm of the state"""
        if self._pure:
            return sqrt(np.vdot(self._state, self._state).real)
        return ops.trace(self._state, self._num_modes)

    def alloc(self, n=1):
        """allocate a number of modes at the end of the state."""
        # base_shape = [self._trunc for i in range(n)]
        if self._pure:
            vac = ops.vacuumState(n, self._trunc)
        else:
            vac = ops.vacuumStateMixed(n, self._trunc)

        self._state = ops.tensor(self._state, vac, self._num_modes, self._pure)
        self._num_modes = self._num_modes + n

    def dealloc(self, modes):
        """Traces out and deallocates the modes in `modes`"""
        if self._pure:
            self._state = ops.mix(self._state, self._num_modes)
            self._pure = False

        self._state = ops.partial_trace(self._state, self._num_modes, modes)
        self._num_modes = self._num_modes - len(modes)

    def prepare_multimode(self, state, modes):
        r"""
        Prepares a given mode or list of modes in the given state.

        After the preparation the system is in a mixed product state,
        with the specified modes replaced by state.
        The given state can be either in tensor form or in matrix/vector form.
        If modes is not ordered, the subsystems of the input are
        reordered to reflect that, i.e., if modes=[3,1], then the first mode
        of state ends up in mode 3 and the second mode of state ends up in
        mode 1 of the output state.
        If modes is None, it is attempted to prepare state in all modes.
        The reduced state on all other modes remains unchanged and
        the final state is product with respect to the partition into
        the modes in modes and the complement.

        Args:
            state (array): vector, matrix, or tensor representation of the ket state or dm state in the fock basis to prepare
            modes (list[int] or non-negative int): The mode(s) into which state is to be prepared. Needs not be ordered.
        """
        if isinstance(modes, int):
            modes = [modes]

        n_modes = len(modes)
        pure_shape = tuple([self._trunc] * n_modes)
        mixed_shape = tuple([self._trunc] * (2 * n_modes))
        pure_shape_as_vector = tuple([self._trunc ** n_modes])
        mixed_shape_as_matrix = tuple([self._trunc ** n_modes] * 2)

        # Do consistency checks
        if self._checks:
            if (
                state.shape != pure_shape
                and state.shape != mixed_shape
                and state.shape != pure_shape_as_vector
                and state.shape != mixed_shape_as_matrix
            ):
                raise ValueError("Incorrect shape for state preparation")
            if len(modes) != len(set(modes)):
                raise ValueError("The specified modes cannot appear multiple times.")

        # reshape to support input both as tensor and vector/matrix
        if state.shape == pure_shape_as_vector:
            state = state.reshape(pure_shape)
        elif state.shape == mixed_shape_as_matrix:
            state = state.reshape(mixed_shape)

        if self._num_modes == n_modes:
            # Hack for marginally faster state preparation
            self._state = state.astype(ops.def_type)
            self._pure = bool(state.shape == pure_shape)
        else:
            if self._pure:
                self._state = ops.mix(self._state, self._num_modes)
                self._pure = False

            if state.shape == pure_shape:
                state = ops.mix(state, len(modes))

            # Take the partial trace
            # todo: For performance the partial trace could be done directly from the pure state. This would of course require a better partial trace function...
            reduced_state = ops.partial_trace(self._state, self._num_modes, modes)

            # Insert state at the end (I know there is also tensor() from ops but it has extra aguments wich only confuse here)
            self._state = np.tensordot(reduced_state, state, axes=0)

            # unless the preparation was meant to go into the last modes in the standard order, we need to swap indices around
        if modes != list(range(self._num_modes - len(modes), self._num_modes)):
            mode_permutation = [x for x in range(self._num_modes) if x not in modes] + modes
            if self._pure:
                scale = 1
                index_permutation = mode_permutation
            else:
                scale = 2
                index_permutation = [
                    scale * x + i for x in mode_permutation for i in (0, 1)
                ]  # two indices per mode if we have pure states
            index_permutation = np.argsort(index_permutation)

            self._state = np.transpose(self._state, index_permutation)

    def prepare(self, state, mode):
        r"""
        Prepares a given mode in a given state.

        This is a simple wrapper for prepare_multimode(), see there for more details.

        Args:
            state (array): vector, matrix, or tensor representation of the ket state or dm state in the fock basis to prepare
            modes (list[int] or non-negative int or None): The mode(s) into which state is to be prepared. Needs not be ordered.
        """
        if isinstance(mode, int):
            mode = [mode]
        self.prepare_multimode(state, mode)

    def prepare_mode_fock(self, n, mode):
        """
        Prepares a mode in a fock state.
        """

        if self._pure:
            self.prepare(ops.fockState(n, self._trunc), mode)
        else:
            st = ops.fockState(n, self._trunc)
            self.prepare(np.outer(st, st.conjugate()), mode)

    def prepare_mode_coherent(self, alpha, mode):
        """
        Prepares a mode in a coherent state.
        """
        if self._pure:
            self.prepare(ops.coherentState(alpha, self._trunc), mode)
        else:
            st = ops.coherentState(alpha, self._trunc)
            self.prepare(np.outer(st, st.conjugate()), mode)

    def prepare_mode_squeezed(self, r, theta, mode):
        """
        Prepares a mode in a squeezed state.
        """
        if self._pure:
            self.prepare(ops.squeezedState(r, theta, self._trunc), mode)
        else:
            st = ops.squeezedState(r, theta, self._trunc)
            self.prepare(np.outer(st, st.conjugate()), mode)

    def prepare_mode_displaced_squeezed(self, alpha, r, phi, mode):
        """
        Prepares a mode in a displaced squeezed state.
        """
        if self._pure:
            self.prepare(ops.displacedSqueezed(alpha, r, phi, self._trunc), mode)
        else:
            st = ops.displacedSqueezed(alpha, r, phi, self._trunc)
            self.prepare(np.outer(st, st.conjugate()), mode)

    def prepare_mode_thermal(self, nbar, mode):
        """
        Prepares a mode in a thermal state.
        """
        st = ops.thermalState(nbar, self._trunc)
        self.prepare(st, mode)

    def phase_shift(self, theta, mode):
        """
        Applies a phase shifter.
        """
        mat = ops.phase(theta, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode])

    def displacement(self, alpha, mode):
        """
        Applies a displacement gate.
        """
        mat = ops.displacement(alpha, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode])

    def beamsplitter(self, t, r, phi, mode1, mode2):
        """
        Applies a beamsplitter.
        """
        mat = ops.beamsplitter(t, r, phi, self._trunc)

        modes = List()
        modes.append(mode1)
        modes.append(mode2)

        self._state = self.apply_twomode_gate(mat, modes, gate="BSgate")

    def squeeze(self, r, theta, mode):
        """
        Applies a squeezing gate.
        """
        mat = ops.squeezing(r, theta, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode])

    def two_mode_squeeze(self, r, theta, mode1, mode2):
        """
        Applies a two-mode squeezing gate.
        """
        mat = ops.two_mode_squeezing(r, theta, self._trunc)

        modes = List()
        modes.append(mode1)
        modes.append(mode2)

        self._state = self.apply_twomode_gate(mat, modes, gate="S2gate")

    def kerr_interaction(self, kappa, mode):
        """
        Applies a Kerr interaction gate.
        """
        mat = ops.kerr(kappa, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode])

    def cross_kerr_interaction(self, kappa, mode1, mode2):
        """
        Applies a cross-Kerr interaction gate.
        """
        mat = ops.cross_kerr(kappa, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode1, mode2])

    def cubic_phase_shift(self, gamma, mode):
        """
        Applies a cubic phase shift gate.
        """
        mat = ops.cubicPhase(gamma, self._hbar, self._trunc)
        self._state = self.apply_gate_BLAS(mat, [mode])

    def is_vacuum(self, tol):
        """
        Tests whether the system is in the vacuum state.
        """
        # fid = <0|\rho|0>
        if self._pure:
            fid = np.abs(self._state.flat[0]) ** 2
        else:
            fid = self._state.flat[0]
        return np.abs(fid - 1) <= tol

    def get_state(self):
        """
        Returns the state of the system in the fock basis along with its purity.
        """
        return self._state, self._pure

    def loss(self, T, mode):
        """
        Applies a loss channel to the state.
        """
        self._apply_channel(ops.lossChannel(T, self._trunc), [mode])

    def measure_fock(self, modes, select=None):
        """
        Measures a list of modes.
        """
        # pylint: disable=singleton-comparison
        if select is not None and np.any(np.array(select) == None):
            raise NotImplementedError("Post-selection lists must only contain numerical values.")

        # Make sure the state is mixed
        if self._pure:
            state = ops.mix(self._state, self._num_modes)
        else:
            state = self._state

        if select is not None:
            # perform post-selection

            # make sure modes and select are the same length
            if len(select) != len(modes):
                raise ValueError(
                    "When performing post-selection, the number of "
                    "selected values (including None) must match the number of measured modes"
                )

            # make sure the select values are all integers or nones
            if not all(isinstance(s, int) or s is None for s in select):
                raise TypeError("The post-select list elements either be integers or None")

            # modes to measure
            measure = [i for i, s in zip(modes, select) if s is None]

            # modes already post-selected:
            selected = [i for i, s in zip(modes, select) if s is not None]
            select_values = [s for s in select if s is not None]

            # project out postselected modes
            self._state = ops.project_reset(
                selected, select_values, self._state, self._pure, self._num_modes, self._trunc
            )

            if self.norm() == 0:
                raise ZeroDivisionError("Measurement has zero probability.")

            self._state = self._state / self.norm()

        else:
            # no post-selection; modes to measure are the modes provided
            measure = modes

        if len(measure) > 0:
            # sampling needs to be performed
            # Compute distribution by tracing out modes not measured, then computing the diagonal
            unmeasured = [i for i in range(self._num_modes) if i not in measure]
            reduced = ops.partial_trace(state, self._num_modes, unmeasured)
            dist = np.ravel(ops.diagonal(reduced, len(measure)).real)
            # kill spurious tiny values (which are sometimes negative)
            dist = dist * ~np.isclose(dist, 0.0)

            # Make a random choice
            if sum(dist) != 1:
                # WARNING: distribution is not normalized, could hide errors
                i = np.random.choice(list(range(len(dist))), p=dist / sum(dist))
            else:
                i = np.random.choice(list(range(len(dist))), p=dist)

            permuted_outcome = ops.unIndex(i, len(measure), self._trunc)

            # Permute the outcome to match the order of the modes in 'measure'
            permutation = np.argsort(measure)
            outcome = [0] * len(measure)
            for i in range(len(measure)):
                outcome[permutation[i]] = permuted_outcome[i]

            # Project the state onto the measurement outcome & reset in vacuum
            self._state = ops.project_reset(
                measure, outcome, self._state, self._pure, self._num_modes, self._trunc
            )

            if self.norm() == 0:
                raise ZeroDivisionError("Measurement has zero probability.")

            self._state = self._state / self.norm()

        # include post-selected values in measurement outcomes
        if select is not None:
            outcome = copy.copy(select)

        return outcome

    def measure_homodyne(self, phi, mode, select=None, **kwargs):
        """
        Performs a homodyne measurement on a mode.
        """
        m_omega_over_hbar = 1 / self._hbar

        # Make sure the state is mixed for reduced density matrix
        if self._pure:
            state = ops.mix(self._state, self._num_modes)
        else:
            state = self._state

        if select is not None:
            meas_result = select
            if isinstance(meas_result, numbers.Number):
                homodyne_sample = float(meas_result)
            else:
                raise TypeError("Selected measurement result must be of numeric type.")
        else:
            # Compute reduced density matrix
            unmeasured = [i for i in range(self._num_modes) if not i == mode]
            reduced = ops.partial_trace(state, self._num_modes, unmeasured)

            # Rotate to measurement basis
            reduced = self.apply_gate_BLAS(
                ops.phase(-phi, self._trunc), [0], state=reduced, pure=False, n=1
            )

            # Create pdf. Same as tf implementation, but using
            # the recursive relation H_0(x) = 1, H_1(x) = 2x, H_{n+1}(x) = 2xH_n(x) - 2nH_{n-1}(x)
            q_mag = kwargs.get("max", 10)
            num_bins = kwargs.get("num_bins", 100000)

            q_tensor, Hvals = ops.hermiteVals(q_mag, num_bins, m_omega_over_hbar, self._trunc)
            H_matrix = np.zeros((self._trunc, self._trunc, num_bins))
            for n, m in product(range(self._trunc), repeat=2):
                H_matrix[n][m] = 1 / sqrt(2 ** n * bang(n) * 2 ** m * bang(m)) * Hvals[n] * Hvals[m]
            H_terms = np.expand_dims(reduced, -1) * np.expand_dims(H_matrix, 0)
            rho_dist = (
                np.sum(H_terms, axis=(1, 2))
                * (m_omega_over_hbar / pi) ** 0.5
                * np.exp(-m_omega_over_hbar * q_tensor ** 2)
                * (q_tensor[1] - q_tensor[0])
            )  # Delta_q for normalization (only works if the bins are equally spaced)

            # Sample from rho_dist. This is a bit different from tensorflow due to how
            # numpy treats multinomial sampling. In particular, numpy returns a
            # histogram of the samples whereas tensorflow gives the list of samples.
            # Numpy also does not use the log probabilities
            probs = rho_dist.flatten().real
            probs /= np.sum(probs)
            sample_hist = np.random.multinomial(1, probs)
            sample_idx = list(sample_hist).index(1)
            homodyne_sample = q_tensor[sample_idx]

        # Project remaining modes into the conditional state
        inf_squeezed_vac = np.array(
            [
                (-0.5) ** (n // 2) * sqrt(bang(n)) / bang(n // 2) if n % 2 == 0 else 0.0 + 0.0j
                for n in range(self._trunc)
            ],
            dtype=ops.def_type,
        )
        alpha = homodyne_sample * sqrt(m_omega_over_hbar / 2)

        composed = np.dot(ops.phase(phi, self._trunc), ops.displacement(alpha, self._trunc))
        eigenstate = self.apply_gate_BLAS(composed, [0], state=inf_squeezed_vac, pure=True, n=1)

        vac_state = np.array(
            [1.0 + 0.0j if i == 0 else 0.0 + 0.0j for i in range(self._trunc)], dtype=ops.def_type
        )
        projector = np.outer(vac_state, eigenstate.conj())

        self._state = self.apply_gate_BLAS(projector, [mode])

        # Normalize
        self._state = self._state / self.norm()

        return homodyne_sample



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Default parameters, environment variables, fixtures, and common routines for the unit tests.
"""
# pylint: disable=redefined-outer-name
import os
import pytest

import strawberryfields as sf
from strawberryfields.engine import LocalEngine
from strawberryfields.program import Program
from strawberryfields.backends.base import BaseBackend
from strawberryfields.backends.fockbackend import FockBackend
from strawberryfields.backends.gaussianbackend import GaussianBackend


try:
    import tensorflow as tf
except (ImportError, ModuleNotFoundError) as e:
    tf_available = False
    tf_version = False
else:
    tf_available = True
    tf_version = tf.__version__


backend_params = [
    pytest.param(FockBackend, marks=pytest.mark.fock),
    pytest.param(GaussianBackend, marks=pytest.mark.gaussian),
]


eng_backend_params = [
    pytest.param("fock", marks=pytest.mark.fock),
    pytest.param("gaussian", marks=pytest.mark.gaussian),
]

if tf_available and tf.__version__[:3] == "1.3":
    from strawberryfields.backends.tfbackend import TFBackend

    backend_params.append(pytest.param(TFBackend, marks=pytest.mark.tf))
    eng_backend_params.append(pytest.param("tf", marks=pytest.mark.tf))
else:
    tf_available = False


# defaults
TOL = 1e-3
CUTOFF = 6
ALPHA = 0.1
HBAR = 1.7
BATCHED = False
BATCHSIZE = 2


@pytest.fixture(scope="session")
def tol():
    """Numerical tolerance for equality tests."""
    return float(os.environ.get("TOL", TOL))


@pytest.fixture(scope="session")
def cutoff():
    """Fock state cutoff"""
    return int(os.environ.get("CUTOFF", CUTOFF))


@pytest.fixture(scope="session")
def alpha():
    """Maximum magnitude of coherent states used in tests"""
    return float(os.environ.get("ALPHA", ALPHA))


@pytest.fixture(scope="session")
def hbar():
    """The value of hbar"""
    sf.hbar = float(os.environ.get("HBAR", HBAR))
    return sf.hbar


@pytest.fixture(
    params=[
        pytest.param(True, marks=pytest.mark.pure),
        pytest.param(False, marks=pytest.mark.mixed),
    ]
)
def pure(request):
    """Whether to run the backend in pure or mixed state mode"""
    return request.param  # bool(int(os.environ.get("PURE", PURE)))


@pytest.fixture(scope="session")
def batch_size():
    """Whether to run the backend in batched mode"""
    if "BATCHSIZE" in os.environ:
        # if user-specified BATCHSIZE provided, then batching is assumed (even if BATCHED=0)
        return int(os.environ["BATCHSIZE"])

    # check if batching is turned on
    batched = bool(int(os.environ.get("BATCHED", BATCHED)))

    if batched:
        # use the default batch size
        return BATCHSIZE

    return None  # no batching


@pytest.fixture
def backend(monkeypatch):
    """Create a mocked out backend fixture for front-end only tests"""
    dummy_backend = BaseBackend()
    with monkeypatch.context() as m:
        # mock out the base backend
        m.setattr(dummy_backend, "add_mode", lambda n: None)
        m.setattr(dummy_backend, "del_mode", lambda n: None)
        m.setattr(dummy_backend, "displacement", lambda alpha, modes: None)
        m.setattr(dummy_backend, "squeeze", lambda r, modes: None)
        m.setattr(dummy_backend, "rotation", lambda r, modes: None)
        m.setattr(dummy_backend, "beamsplitter", lambda t, r, m1, m2: None)
        m.setattr(dummy_backend, "measure_homodyne", lambda phi, modes, select, shots: 5)
        m.setattr(dummy_backend, "state", lambda modes, shots: None)
        m.setattr(dummy_backend, "reset", lambda: None)
        dummy_backend.get_cutoff_dim = lambda: 6
        yield dummy_backend


@pytest.fixture(scope="session")
def print_fixtures(cutoff, hbar, batch_size):
    """Print the test configuration at the beginning of the session"""
    print(
        "FIXTURES: cutoff = {}, hbar = {}, batch_size = {}".format(
            cutoff, hbar, pure, batch_size
        )
    )


@pytest.fixture(params=backend_params)
def setup_backend(
    request, cutoff, pure, batch_size
):  # pylint: disable=redefined-outer-name
    """Parameterized fixture, used to automatically create a backend of certain number of modes.

    This fixture should only be used in backend tests, as it bypasses Engine and
    initializes the backend directly.

    Every test that uses this fixture, or a fixture that depends on it
    will be run three times, one for each backend.

    To explicitly mark a test to only work on certain backends, you may
    use the ``@pytest.mark.backends()`` fixture. For example, for a test that
    only works on the TF and Fock backends, ``@pytest.mark.backends('tf', 'fock').
    """

    def _setup_backend(num_subsystems):
        """Factory function"""
        backend = request.param()
        backend.begin_circuit(
            num_subsystems=num_subsystems,
            cutoff_dim=cutoff,
            pure=pure,
            batch_size=batch_size,
        )
        return backend

    return _setup_backend


@pytest.fixture(params=eng_backend_params)
def setup_backend_pars(
    request, cutoff, pure, batch_size
):  # pylint: disable=redefined-outer-name
    """Parameterized fixture, a container for the backend parameters.

    Every test that uses this fixture, or a fixture that depends on it
    will be run three times, one for each backend.

    To explicitly mark a test to only work on certain backends, you may
    use the ``@pytest.mark.backends()`` fixture. For example, for a test that
    only works on the TF and Fock backends, ``@pytest.mark.backends('tf', 'fock').
    """
    return request.param, {"cutoff_dim": cutoff, "pure": pure, "batch_size": batch_size}


@pytest.fixture
def setup_eng(setup_backend_pars):  # pylint: disable=redefined-outer-name
    """Parameterized fixture, used to automatically create an Engine and a Program with a certain number of modes."""

    def _setup_eng(num_subsystems, **kwargs):
        """Factory function"""
        prog = Program(num_subsystems)
        backend, backend_options = setup_backend_pars
        backend_options.update(kwargs)  # override defaults with kwargs
        eng = LocalEngine(backend=backend, backend_options=backend_options)
        return eng, prog

    return _setup_eng

def pytest_runtest_setup(item):
    """Automatically skip tests if they are marked for only certain backends"""
    if tf_available:
        allowed_backends = {"gaussian", "tf", "fock"}
    else:
        allowed_backends = {"gaussian", "fock"}

    # load the marker specifying what the backend is
    marks = {mark.name for mark in item.iter_markers() if mark.name in allowed_backends}

    # load a marker specifying whether the test only works with certain backends
    test_backends = [mark.args for mark in item.iter_markers(name="backends")]

    if not test_backends:
        # if the test hasn't specified that it runs on particular backends,
        # assume it will work with all backends
        test_backends = allowed_backends
    else:
        # otherwise, extract the set of backend strings the test works with
        test_backends = test_backends[0]

    for b in marks:
        if b not in test_backends:
            pytest.skip(
                "\nTest {} only runs with {} backend(s), "
                "but {} backend provided".format(item.nodeid, test_backends, b)
            )

    # skip broken tests
    for mark in item.iter_markers(name="broken"):
        if mark.args:
            pytest.skip("Broken test skipped: {}".format(*mark.args))
        else:
            pytest.skip("Test skipped as corresponding code base is currently broken!")



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gaussian operations"""

import numpy as np

from scipy.linalg import sqrtm

from thewalrus.quantum import (
    density_matrix_element,
    is_pure_cov,
    pure_state_amplitude,
    state_vector,
    density_matrix,
)


def fock_amplitudes_one_mode(alpha, cov, cutoff):
    """ Returns the Fock space density matrix of gaussian state characterized
    by a complex displacement alpha and a (symmetric) covariance matrix
    The Fock ladder ladder goes from 0 to cutoff-1"""
    r = 2 * np.array([alpha.real, alpha.imag])
    if is_pure_cov(cov):
        psi = state_vector(r, cov, normalize=True, cutoff=cutoff)
        rho = np.outer(psi, psi.conj())
        return rho
    return density_matrix(r, cov, normalize=True, cutoff=cutoff)


def sm_fidelity(mu1, mu2, cov1, cov2, tol=1e-8):
    """ Calculates the squared fidelity between the gaussian states s1 and s2. It uses the formulas from
    Quantum Fidelity for Arbitrary Gaussian States
    Leonardo Banchi, Samuel L. Braunstein, and Stefano Pirandola
    Phys. Rev. Lett. 115, 260501 – Published 22 December 2015
    The function returns the square of the quantity defined in the reference cited above.
    Note however that our matrices need to be multiplied by 1/2 to get theirs and our vectors
    need to be divided by sqrt(1/2) equivalently the factor in the exponential is not multiplied
    by 2*1/4 but instead by 2*1/8=0.25
    """
    # pylint: disable=duplicate-code
    v1 = 0.5 * cov1
    v2 = 0.5 * cov2
    deltar = mu1 - mu2
    n = 1
    W = omega(2 * n)

    si12 = np.linalg.inv(v1 + v2)
    vaux = np.dot(np.dot(np.transpose(W), si12), 0.25 * W + np.dot(v2, np.dot(W, v1)))

    p1 = np.dot(vaux, W)
    p1 = np.dot(p1, p1)
    p1 = np.identity(2 * n) + 0.25 * np.linalg.inv(p1)
    if np.linalg.norm(p1) < tol:
        p1 = np.zeros_like(p1)
    else:
        p1 = sqrtm(p1)
    p1 = 2 * (p1 + np.identity(2 * n))
    p1 = np.dot(p1, vaux).real
    f = np.sqrt(np.linalg.det(si12) * np.linalg.det(p1)) * np.exp(
        -0.25 * np.dot(np.dot(deltar, si12), deltar).real
    )
    return f


def chop_in_blocks(m, idtodelete):
    """
    Splits a (symmetric) matrix into 3 blocks, A, B, C
    Blocks A and B are diagonal blocks and C is the offdiagonal block
    idtodelete specifies which indices go into B.
    """
    A = np.copy(m)
    A = np.delete(A, idtodelete, axis=0)
    A = np.delete(A, idtodelete, axis=1)
    B = np.delete(m[:, idtodelete], idtodelete, axis=0)
    C = np.empty((len(idtodelete), (len(idtodelete))))
    for localindex, globalindex in enumerate(idtodelete):
        for localindex1, globalindex1 in enumerate(idtodelete):
            C[localindex, localindex1] = m[globalindex, globalindex1]
    return (A, B, C)


def chop_in_blocks_vector(v, idtodelete):
    """
    Splits a vector into two vectors, where idtodelete specifies
    which elements go into vb
    """
    idtokeep = list(set(np.arange(len(v))) - set(idtodelete))
    va = v[idtokeep]
    vb = v[idtodelete]
    return (va, vb)


def reassemble(A, idtodelete):
    """
    Puts the matrix A inside a larger matrix of dimensions
    dim(A)+len(idtodelete)
    The empty space are filled with zeros (offdiagonal) and ones (diagonals)
    """
    ntot = len(A) + len(idtodelete)
    ind = set(np.arange(ntot)) - set(idtodelete)
    newmat = np.zeros((ntot, ntot))
    for i, i1 in enumerate(ind):
        for j, j1 in enumerate(ind):
            newmat[i1, j1] = A[i, j]

    for i in idtodelete:
        newmat[i, i] = 1.0
    return newmat


def reassemble_vector(va, idtodelete):
    r"""Creates a vector with zeros indices idtodelete
    and everywhere else it puts the entries of va
    """
    ntot = len(va) + len(idtodelete)
    ind = set(np.arange(ntot)) - set(idtodelete)
    newv = np.zeros(ntot)
    for j, j1 in enumerate(ind):
        newv[j1] = va[j]
    return newv


def omega(n):
    """ Utility function to calculate fidelities"""
    x = np.zeros(n)
    x[0::2] = 1
    A = np.diag(x[0:-1], 1)
    W = A - np.transpose(A)
    return W


def xmat(n):
    """ Returns the matrix ((0, I_n), (I, 0_n))"""
    idm = np.identity(n)
    return np.concatenate(
        (np.concatenate((0 * idm, idm), axis=1), np.concatenate((idm, 0 * idm), axis=1)), axis=0
    ).real


def fock_prob(mu, cov, ocp):
    """
    Calculates the probability of measuring the gaussian state s2 in the photon number
    occupation pattern ocp"""
    if is_pure_cov(cov):
        return np.abs(pure_state_amplitude(mu, cov, ocp, check_purity=False)) ** 2

    return density_matrix_element(mu, cov, list(ocp), list(ocp)).real



# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package contains the modules that make up the
Strawberry Fields backends. This includes photonic simulators,
shared numerical operations, and states objects returned by
statevector simulators.

Local statevector simulators
----------------------------

Below are all available local statevector backends currently
provided by Strawberry Fields. These simulators all run locally,
provide access to the state after simulation, and the state is
preserved between engine runs.

.. currentmodule:: strawberryfields.backends
.. autosummary::
    :toctree: api

    FockBackend
    GaussianBackend
    ~tfbackend.TFBackend

.. raw:: html

    <div style="display: none;">

.. currentmodule:: strawberryfields.backends
.. autosummary::
    :toctree: api

    BaseFockState
    ~gaussianbackend.states.GaussianState
    ~tfbackend.states.FockStateTF

.. raw:: html

    </div>

Backend API
-----------

A list of the abstract base classes that define the
statevector backend API

.. currentmodule:: strawberryfields.backends
.. autosummary::
    :toctree: api

    BaseState
    BaseGaussianState
    BaseBackend
    BaseFock
    BaseGaussian

Utility modules
---------------

The following utility modules are provided for
backend development.

.. currentmodule:: strawberryfields.backends
.. autosummary::
    :toctree: api

    shared_ops
"""

from .base import BaseBackend, BaseFock, BaseGaussian, ModeMap
from .gaussianbackend import GaussianBackend
from .fockbackend import FockBackend
from .states import BaseState, BaseGaussianState, BaseFockState


__all__ = [
    "BaseBackend",
    "BaseFock",
    "BaseGaussian",
    "FockBackend",
    "GaussianBackend",
    "TFBackend",
    "BaseState",
    "BaseFockState",
    "BaseGaussianState",
]


virtual_backends = ["X8_01"]

local_backends = {b.short_name: b for b in (BaseBackend, GaussianBackend, FockBackend)}


def load_backend(name):
    """Loads the specified backend by mapping a string
    to the backend type, via the ``local_backends``
    dictionary. Note that this function is used by the
    frontend only, and should not be user-facing.
    """
    if name == "tf":
        # treat the tensorflow backend differently, to
        # isolate the import of tensorflow
        from .tfbackend import TFBackend

        return TFBackend()

    if name in virtual_backends:
        # Backend is a remote device/simulator, that has a
        # defined circuit spec, but no local backend class.
        # By convention, the short name and corresponding
        # circuit spec are the same.
        backend_attrs = {"short_name": name, "circuit_spec": name}
        backend_class = type(name, (BaseBackend,), backend_attrs)
        return backend_class()

    if name in local_backends:
        backend = local_backends[name]()
        return backend

    raise ValueError("Backend '{}' is not supported.".format(name))
